import copy
import hashlib
import warnings

import numpy as np
import pandas as pd
from typing import Any

from mirp._images.generic_image import GenericImage
from mirp.settings.generic import SettingsClass


class MaskImage(GenericImage):

    def __init__(self, **kwargs):
        # Declare local attributes first because these may be updated in super()__init__.
        self.image_encoded = False
        self.alteration_size: float = 0.0
        self.slic_randomisation_id: int | None = None

        super().__init__(**kwargs)

    def is_empty(self):
        if self.image_encoded:
            return False
        else:
            return self.image_data is None

    def is_empty_mask(self):
        if self.is_empty():
            return True

        if np.sum(self.get_voxel_grid()) == 0:
            return True

        return False

    def encode_voxel_grid(self):
        # Check if image data are present, or are already encoded.
        if self.is_empty() or self.image_encoded is True:
            return

        # Check that the image consists of boolean values.
        if self.image_data.dtype != bool:
            return

        rle_end = np.array(np.append(
            np.where(self.image_data.ravel()[1:] != self.image_data.ravel()[:-1]), np.prod(self.image_dimension) - 1))
        rle_start = np.cumsum(np.append(0, np.diff(np.append(-1, rle_end))))[:-1]
        rle_val = self.image_data.ravel()[rle_start]

        # Check whether the image mask is empty (consists of 0s)
        if np.all(~rle_val):
            self.image_data = None
            self.image_encoded = True
        else:
            # Select only masked parts of the image for further compression
            rle_start = rle_start[rle_val]
            rle_end = rle_end[rle_val]

            # Create zip
            self.image_data = zip(rle_start, rle_end)
            self.image_encoded = True

    def decode_voxel_grid(self, in_place=True):
        # Check if the voxel grid is already decoded.
        if not self.image_encoded:
            if in_place:
                return
            else:
                return self.image_data

        decoded_voxel_grid = np.zeros(np.prod(self.image_dimension), dtype=bool)

        # Check if the image contains masked regions, and unzip them.
        if self.image_data is not None:
            decoding_zip = copy.deepcopy(self.image_data)
            for ii, jj in decoding_zip:
                decoded_voxel_grid[ii:jj + 1] = True

        # Restore shape to original grid.
        decoded_voxel = decoded_voxel_grid.reshape(self.image_dimension)

        if in_place:
            self.image_data = decoded_voxel
            self.image_encoded = False

        else:
            return decoded_voxel

    def set_voxel_grid(self, voxel_grid: np.ndarray):
        self.image_encoded = False
        super().set_voxel_grid(voxel_grid=voxel_grid)

        # Force encoding if the data are boolean.
        if voxel_grid.dtype == bool:
            self.encode_voxel_grid()

    def get_voxel_grid(self) -> None | np.ndarray:
        return self.decode_voxel_grid(in_place=False)

    def update_image_data(self):
        # Do not update image data if the data are absent, it is encoded, or the image is already a boolean mask.
        if self.is_empty() or self.image_encoded or self.image_data.dtype == bool:
            return

        # Ensure that mask consists of boolean values.
        self.image_data = np.around(self.image_data, 6) >= np.around(0.5, 6)
        self.encode_voxel_grid()

    @staticmethod
    def get_interpolation_spline_order(settings: SettingsClass):
        return settings.roi_interpolate.spline_order

    def add_noise(self, **kwargs):
        pass

    def saturate(self, **kwargs):
        pass

    def normalise_intensities(self, **kwargs):
        pass

    def crop(
            self,
            ind_ext_z=None,
            ind_ext_y=None,
            ind_ext_x=None,
            xy_only=False,
            z_only=False):

        if self.image_encoded:
            self.decode_voxel_grid()
            super().crop(
                ind_ext_x=ind_ext_x,
                ind_ext_y=ind_ext_y,
                ind_ext_z=ind_ext_z,
                xy_only=xy_only,
                z_only=z_only
            )
            self.encode_voxel_grid()

        else:
            super().crop(
                ind_ext_x=ind_ext_x,
                ind_ext_y=ind_ext_y,
                ind_ext_z=ind_ext_z,
                xy_only=xy_only,
                z_only=z_only
            )

    def crop_to_size(self, center, crop_size):

        if self.image_encoded:
            self.decode_voxel_grid()
            super().crop_to_size(
                center=center,
                crop_size=crop_size
            )
            self.encode_voxel_grid()

        else:
            super().crop_to_size(
                center=center,
                crop_size=crop_size
            )

    def dilate(
            self,
            distance: float | None = None,
            voxel_distance: float | None = None
    ):
        from mirp._features.utilities import rep
        import scipy.ndimage as ndi

        # Skip if the mask does not exist
        if self.is_empty():
            return

        # Check if any distance is provided for dilation
        if voxel_distance is None and distance is None:
            return

        # Check whether voxel are isotropic.
        if not self.is_isotropic():
            warnings.warn(
                "Non-uniform voxel spacing was detected. Mask dilation requires uniform voxel spacing.", UserWarning
            )

        # Set spacing
        if self.separate_slices:
            spacing = np.array(self.image_spacing)[[1, 2]]
        else:
            spacing = np.array(self.image_spacing)

        # Derive filter extension and distance
        if distance is not None:
            base_ext: int = np.max([np.floor(distance / np.max(spacing)).astype(int), 0])
        else:
            base_ext: int = int(voxel_distance)
            distance = voxel_distance * np.max(spacing)

        # Check that the dilation size is larger than 1, i.e. base_ext > 0.
        if base_ext <= 0:
            warnings.warn(
                f"Mask was not dilated as the distance ({distance}) was too small compared to voxel spacing ("
                f"{np.max(spacing)})", UserWarning
            )
            return

        # Skip if there is no mask to dilate.
        if self.is_empty_mask():
            return

        # Create displacement map
        df_base = pd.DataFrame({
            "x": rep(
                x=np.arange(-base_ext, base_ext + 1),
                each=(2 * base_ext + 1) * (2 * base_ext + 1),
                times=1),
            "y": rep(
                x=np.arange(-base_ext, base_ext + 1),
                each=2 * base_ext + 1,
                times=2 * base_ext + 1),
            "z": rep(
                x=np.arange(-base_ext, base_ext + 1),
                each=1,
                times=(2 * base_ext + 1) * (2 * base_ext + 1))
        })

        # Calculate distances for displacement map.
        df_base["dist"] = np.sqrt(
            np.sum(np.multiply(df_base.loc[:, ("z", "y", "x")].values, self.image_spacing) ** 2.0, axis=1))

        # Identify elements in range.
        if self.separate_slices:
            df_base["in_range"] = np.logical_and(df_base.dist <= distance, df_base.z == 0)
        else:
            df_base["in_range"] = df_base.dist <= distance

        # Update voxel coordinates to start at [0,0,0].
        df_base["x"] += base_ext
        df_base["y"] += base_ext
        df_base["z"] += base_ext

        # Generate geometric filter structure.
        geom_struct = np.zeros(
            shape=(np.max(df_base.z) + 1, np.max(df_base.y) + 1, np.max(df_base.x) + 1),
            dtype=bool
        )
        geom_struct[df_base.z.astype(int), df_base.y.astype(int), df_base.x.astype(int)] = df_base.in_range

        # Dilate roi mask and store voxel grid.
        self.set_voxel_grid(voxel_grid=ndi.binary_dilation(
            self.get_voxel_grid(),
            structure=geom_struct,
            iterations=1)
        )

        self.alteration_size = distance

    def erode(
            self,
            max_eroded_volume_fraction: float = 0.8,
            distance: float | None = None,
            voxel_distance: float | None = None
    ):
        import scipy.ndimage as ndi

        # Skip if the mask does not exist
        if self.is_empty():
            return

        # Check if any distance is provided for dilation
        if voxel_distance is None and distance is None:
            return

        # Check whether voxel are isotropic.
        if not self.is_isotropic():
            warnings.warn(
                "Non-uniform voxel spacing was detected. Mask erosion requires uniform voxel spacing.", UserWarning
            )

        # Set spacing.
        if self.separate_slices:
            spacing = np.array(self.image_spacing)[[1, 2]]
        else:
            spacing = np.array(self.image_spacing)

        # Set geometrical structure. For 2D, the structures in different slices are set to 0.
        geom_struct = ndi.generate_binary_structure(3, 1)
        if self.separate_slices:
            geom_struct[(0, 2), :, :] = False

        # Set number of erosion steps.
        if voxel_distance is None:
            erode_steps = int(np.max([np.round(np.abs(distance) / np.max(spacing)), 0]))
        else:
            erode_steps = int(np.abs(voxel_distance))
            distance = voxel_distance * np.max(spacing)

        # Check that the number of erosion steps is positive.
        if erode_steps <= 0:
            warnings.warn(
                f"Mask was not eroded as the distance ({distance}) was too small compared to voxel spacing ("
                f"{np.max(spacing)})", UserWarning
            )
            return

        # Skip if there is no mask to erode.
        if self.is_empty_mask():
            return

        # Determine initial volume.
        current_mask = previous_mask = self.get_voxel_grid()
        initial_volume = np.sum(previous_mask)

        # Iterate over erosion steps.
        for step in np.arange(0, erode_steps):

            # Perform erosion
            current_mask = ndi.binary_erosion(previous_mask, structure=geom_struct, iterations=1)

            # Calculate volume of the eroded volume
            current_volume = np.sum(current_mask)

            # Stop erosion if the volume shrinks below 80 percent of the original volume due to erosion and voxels
            # from the previous erosion step.
            if current_volume * 1.0 / initial_volume < max_eroded_volume_fraction:
                current_mask = previous_mask
                break
            else:
                previous_mask = current_mask

        # Set updated voxels.
        self.set_voxel_grid(voxel_grid=current_mask)

        self.alteration_size = distance

    def fractional_volume_change(
            self,
            fractional_change: float | None = None
    ):
        import scipy.ndimage as ndi

        # Skip if the roi does not exist
        if self.is_empty():
            return

        if fractional_change is None or np.around(fractional_change, 6) == 0.0:
            return

        # Check whether voxel are isotropic.
        if not self.is_isotropic():
            warnings.warn(
                "Non-uniform voxel spacing was detected. Mask erosion requires uniform voxel spacing.", UserWarning
            )

        # Skip if there is no mask to change.
        if self.is_empty_mask():
            return

        # Set alteration size.
        self.alteration_size = fractional_change

        # Set geometrical structure. For 2D, the structures in different slices are set to 0.
        geom_struct = ndi.generate_binary_structure(3, 1)
        if self.separate_slices:
            geom_struct[(0, 2), :, :] = False

        # Determine original volume
        previous_mask = self.get_voxel_grid()
        initial_volume = np.sum(previous_mask)

        # Iteratively grow or shrink the volume. The loop terminates through break statements
        while True:
            if fractional_change > 0.0:
                current_mask = ndi.binary_dilation(previous_mask, structure=geom_struct, iterations=1)
            else:
                current_mask = ndi.binary_erosion(previous_mask, structure=geom_struct, iterations=1)

            current_volume = np.sum(current_mask)
            if current_volume == 0:
                break

            if 0.0 < fractional_change <= current_volume / initial_volume - 1.0:
                break

            if 0.0 > fractional_change >= current_volume / initial_volume - 1.0:
                break

            # Replace previous mask by the current mask.
            previous_mask = current_mask

        # Start randomiser.
        m = hashlib.sha1(usedforsecurity=False)
        m = self.update_hash(m=m)
        randomiser = np.random.default_rng(int(m.hexdigest(), 16))

        # Randomly add/remove border voxels until desired growth/shrinkage is achieved
        if not current_volume / initial_volume - 1.0 == fractional_change:
            additional_vox = np.abs(int(np.floor(initial_volume * (1.0 + fractional_change) - np.sum(previous_mask))))
            if additional_vox > 0:
                border_voxel_ind = np.array(np.where(np.logical_xor(previous_mask, current_mask)))
                select_ind = randomiser.choice(a=border_voxel_ind.shape[1], size=additional_vox, replace=False)
                border_voxel_ind = border_voxel_ind[:, select_ind]
                if fractional_change > 0.0:
                    previous_mask[border_voxel_ind[0, :], border_voxel_ind[1, :], border_voxel_ind[2, :]] = True
                else:
                    previous_mask[border_voxel_ind[0, :], border_voxel_ind[1, :], border_voxel_ind[2, :]] = False

        # Set the new roi
        self.set_voxel_grid(voxel_grid=previous_mask)

    def randomise_mask(
            self,
            image: GenericImage,
            boundary: float = 25.0,
            repetitions: int = 1,
            intensity_range: tuple[Any, Any] = tuple([np.nan, np.nan])
    ):
        """Use SLIC to randomise the roi based on supervoxels"""
        from scipy.ndimage import binary_closing
        from mirp._image_processing.utilities import set_intensity_range
        from mirp._image_processing.cropping import crop

        # Skip if no randomisation is required.
        if repetitions < 1:
            return None

        # Skip if the roi or image do not exist
        if self.is_empty() or image.is_empty():
            return None

        # Skip if there is no mask to change.
        if self.is_empty_mask():
            return None

        # Crop image and mask to accelerate segmentation process.
        cropped_image, cropped_mask = crop(
            image=image,
            masks=self,
            boundary=boundary,
            xy_only=False,
            z_only=False,
            in_place=False
        )

        # Type hinting.
        cropped_mask: MaskImage = cropped_mask

        # Get supervoxels.
        intensity_range = set_intensity_range(image=image, mask=self, intensity_range=intensity_range)
        image_segments = cropped_image.get_supervoxels(intensity_range=intensity_range)
        overlap_indices, overlap_fractions, overlap_size = cropped_mask.get_supervoxel_overlap(
            image_segments=image_segments
        )

        # Skip if there are no overlapping supervoxels.
        if overlap_indices is None:
            return None

        # Set the highest overlap to 1.0 to ensure selection of at least 1 supervoxel.
        overlap_fractions[np.argmax(overlap_fractions)] = 1.0

        # Always include supervoxels with 90% coverage and always exclude those with less than 20% coverage.
        overlap_fractions[overlap_fractions >= 0.90] = 1.0
        overlap_fractions[overlap_fractions < 0.20] = 0.0

        # Determine grid indices of the resected grid with respect to the original image grid.
        grid_origin = image.to_voxel_coordinates(x=np.array(cropped_image.image_origin))
        grid_origin = grid_origin.astype(int)

        # Initialise list of randomised masks.
        randomised_masks = []

        # Start randomiser.
        m = hashlib.sha1(usedforsecurity=False)
        m = self.update_hash(m=m)
        randomiser = np.random.default_rng(int(m.hexdigest(), 16))

        for ii in range(repetitions):
            # Draw random numbers between 0.0 and 1.0.
            random_inclusion = randomiser.random(size=len(overlap_fractions))

            # Select those segments where the random number is less than the overlap fraction - i.e. the fraction is the
            # probability of selecting the supervoxel.
            included_segments = overlap_indices[np.less(random_inclusion, overlap_fractions)]

            # Replace randomised contour in original roi voxel space.
            new_mask_data = np.zeros(shape=self.image_dimension, dtype=bool)
            new_mask_data[
                grid_origin[0]: grid_origin[0] + cropped_mask.image_dimension[0],
                grid_origin[1]: grid_origin[1] + cropped_mask.image_dimension[1],
                grid_origin[2]: grid_origin[2] + cropped_mask.image_dimension[2]
            ] = np.reshape(np.isin(np.ravel(image_segments), included_segments), cropped_mask.image_dimension)

            # Apply binary closing to close gaps.
            new_mask_data = binary_closing(input=new_mask_data)

            # Set mask and randomisation ID.
            randomised_mask = self.copy(drop_image=True)
            randomised_mask.set_voxel_grid(new_mask_data)
            randomised_mask.slic_randomisation_id = ii
            randomised_masks += [randomised_mask]

        if len(randomised_masks) == 0:
            return None

        return randomised_masks

    def get_supervoxel_overlap(
            self,
            image_segments: np.ndarray | None
    ) -> tuple[np.ndarray | None, np.ndarray | None, np.ndarray | None]:
        """Determines overlap of supervoxels with other the region of interest"""

        # Return None in case image segments and/or mask are missing
        if image_segments is None or self.is_empty() or self.is_empty_mask():
            return None, None, None

        # Determine labels and the voxel count of the masked image.
        overlap_segment_labels, overlap_size = np.unique(
            np.multiply(image_segments, self.get_voxel_grid()),
            return_counts=True
        )

        # Find supervoxels with any overlap with the mask.
        overlap_size = overlap_size[overlap_segment_labels > 0]
        overlap_segment_labels = overlap_segment_labels[overlap_segment_labels > 0]

        if len(overlap_size) == 0:
            return None, None, None

        # Check the actual size of the segments overlapping with the current contour
        full_segment_size = list(map(lambda x: np.sum([image_segments == x]), overlap_segment_labels))

        # Calculate the fraction of overlap
        overlap_fraction = overlap_size / full_segment_size

        return overlap_segment_labels, overlap_fraction, overlap_size

    def get_bounding_box(self):
        if self.is_empty() or self.is_empty_mask():
            return None, None, None

        z_ind, y_ind, x_ind = np.where(self.get_voxel_grid())

        return (
            tuple([np.min(z_ind), np.max(z_ind)]),
            tuple([np.min(y_ind), np.max(y_ind)]),
            tuple([np.min(x_ind), np.max(x_ind)])
        )

    def get_center_position(self) -> list[Any]:
        """Identify location of the geometric center of the roi."""
        # Return a NaN if no roi is present
        if self.is_empty():
            return [np.nan, np.nan, np.nan]

        # Determine indices of voxels included in the roi
        z_ind, y_ind, x_ind = np.where(self.get_voxel_grid())

        return [np.mean(z_ind), np.mean(y_ind), np.mean(x_ind)]

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        # Alteration size.
        if self.alteration_size is not None and not self.alteration_size == 0.0:
            descriptors += ["vol", str(self.alteration_size)]

        # Supervoxel randomisation id
        if self.slic_randomisation_id is not None:
            descriptors += ["svx", str(self.slic_randomisation_id)]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        attributes = super().get_export_attributes()

        # Alteration size.
        if self.alteration_size is not None and not self.alteration_size == 0.0:
            attributes.update({"mask_alteration_size": self.alteration_size})

        # Supervoxel randomisation id
        if self.slic_randomisation_id is not None:
            attributes.update({"mask_randomisation_id": self.slic_randomisation_id})

        return attributes



import warnings

import numpy as np
from pydicom import dcmread

from mirp._data_import.dicom_file import MaskDicomFile
from mirp._data_import.utilities import get_pydicom_meta_tag, has_pydicom_meta_tag
from mirp._masks.base_mask import BaseMask


class MaskDicomFileSEG(MaskDicomFile):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        return False

    def create(self):
        return self

    def _complete_image_origin(self, force=False, frame_id=None):
        return

    def _complete_image_orientation(self, force=False, frame_id=None):
        return

    def _complete_image_spacing(self, force=False, frame_id=None):
        return

    def _complete_image_dimensions(self, force=False):
        return

    def check_mask(self, raise_error=True):
        if self.image_metadata is None:
            raise TypeError("DEV: the image_meta_data attribute has not been set.")

        if not get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0062, 0x0002), test_tag=True):
            if raise_error:
                warnings.warn(
                    f"The current SEG set did not contain any segment sequences. [{self.describe_self()}]",
                )
            return False

        roi_name_present = [
            get_pydicom_meta_tag(
                dcm_seq=current_segment_sequence, tag=(0x0062, 0x0005), tag_type="str", default=None)
            for current_segment_sequence in self.image_metadata[(0x0062, 0x0002)]
        ]

        if len(roi_name_present) == 0 and raise_error:
            warnings.warn(
                f"The current SEG set does not contain any segmentations. [{self.describe_self()}]"
            )
            return False

        if any(x is None for x in roi_name_present) and raise_error:
            warnings.warn(
                f"The current SEG set does not contain any labels. [{self.describe_self()}]"
            )
            return False

        return True

    def load_data(self, **kwargs):
        pass

    def to_object(self, **kwargs) -> None | list[BaseMask]:

        self.load_metadata()
        self.set_object_metadata()
        if not self.check_mask():
            return None

        mask_list = []

        # Find which roi numbers (3006,0022) are associated with which roi names (3004,0024).
        roi_name_present = [
            get_pydicom_meta_tag(
                dcm_seq=current_segment_sequence, tag=(0x0062, 0x0005), tag_type="str", default=None)
            for current_segment_sequence in self.image_metadata[(0x0062, 0x0002)]
        ]
        roi_number_present = [
            get_pydicom_meta_tag(
                dcm_seq=current_segment_sequence, tag=(0x0062, 0x0004), tag_type="int", default=None)
            for current_segment_sequence in self.image_metadata[(0x0062, 0x0002)]
        ]

        # Identify user-provided roi names.
        provided_roi_names = None
        if isinstance(self.roi_name, str):
            provided_roi_names = [self.roi_name]
        elif isinstance(self.roi_name, list):
            provided_roi_names = self.roi_name
        elif isinstance(self.roi_name, dict):
            provided_roi_names = list(self.roi_name.keys())

        if provided_roi_names is not None:
            roi_number_present = [
                x for ii, x in enumerate(roi_number_present) if roi_name_present[ii] in provided_roi_names
            ]
            roi_name_present = [x for x in roi_name_present if x in provided_roi_names]

        if len(roi_number_present) == 0:
            return None

        # Determine which frames corresponds to which roi numbers.
        frame_list = []
        for roi_number in roi_number_present:
            frame_list += [list(self._generate_frame_index(roi_index=roi_number))]

        # Read mask data.
        dcm = dcmread(self.file_path, stop_before_pixels=False, force=True)
        seg_mask_data = dcm.pixel_array

        for ii, roi_number in enumerate(roi_number_present):
            frames = frame_list[ii]

            # Acquire mask spacing
            mask_spacing = self._get_mask_spacing(frames=frames)
            for frame in frames:
                frame_spacing = self._get_mask_spacing(frames=[frame])
                if frame_spacing != mask_spacing:
                    raise ValueError(
                        f"The DICOM SEG file has inconsistent mask spacing [{self.describe_self()}]."
                    )

            # Acquire mask orientation
            mask_orientation = self._get_mask_orientation(frames=frames)
            for frame in frames:
                frame_orientation = self._get_mask_orientation(frames=[frame])
                if not np.array_equal(frame_orientation, mask_orientation):
                    raise ValueError(
                        f"The DICOM SEG file has inconsistent orientation [{self.describe_self()}]."
                    )

            # For each frame, determine slice position relative to the first frame. First determine frame origins. Then
            # Determine slice position
            mask_origin = self._get_mask_origin(frames=frames)
            frame_origins = [
                self._get_mask_origin(frames=[frame]) for frame in frames
            ]
            if all(frame_origin == mask_origin for frame_origin in frame_origins):
                # If origin is not set per frame, but is shared, assume that frames are sequential.
                mask_slice_number = np.arange(len(frames))
            else:
                frame_voxel_origins = [
                    self.to_voxel_coordinates(
                        x=np.array(frame_origin),
                        origin=np.array(mask_origin),
                        orientation=mask_orientation,
                        spacing=np.array(mask_spacing)
                    )
                    for frame_origin in frame_origins
                ]
                mask_slice_number = [int(x[0]) for x in frame_voxel_origins]

            # Determine mask dimensions.
            mask_dimension = tuple([max(mask_slice_number) + 1, seg_mask_data.shape[1], seg_mask_data.shape[2]])

            # Create data and insert frame at the correct slice.
            mask_data = np.zeros(mask_dimension, dtype=seg_mask_data.dtype)
            for jj, frame in enumerate(frames):
                mask_data[mask_slice_number[jj], :, :] = seg_mask_data[frame, :, :]

            current_roi_name = [
                x
                for jj, x in enumerate(roi_name_present)
                if roi_number_present[jj] == roi_number
            ][0]

            if isinstance(self.roi_name, dict):
                current_roi_name = self.roi_name.get(current_roi_name)

            mask_list += [
                BaseMask(
                    roi_name=current_roi_name,
                    sample_name=self.sample_name,
                    image_modality=self.modality,
                    image_data=mask_data,
                    image_spacing=mask_spacing,
                    image_origin=mask_origin,
                    image_orientation=mask_orientation,
                    image_dimensions=mask_dimension,
                    metadata=self.object_metadata
                )
            ]

        if len(mask_list) == 0:
            return None

        return mask_list

    def _generate_frame_index(self, roi_index: int):
        for ii in np.arange(
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0008), tag_type="int")):
            current_roi_number = None
            if has_pydicom_meta_tag(self.image_metadata, tag=(0x5200, 0x9229)):
                if has_pydicom_meta_tag(self.image_metadata[(0x5200, 0x9229)][0], tag=(0x0062, 0x000A)):
                    current_roi_number = get_pydicom_meta_tag(
                        dcm_seq=self.image_metadata[(0x5200, 0x9229)][0][(0x0062, 0x000A)][0],
                        tag=(0x0062, 0x000B),
                        tag_type="int",
                        default=None
                    )

            if current_roi_number is None:
                current_roi_number = get_pydicom_meta_tag(
                    dcm_seq=self.image_metadata[(0x5200, 0x9230)][ii][(0x0062, 0x000A)][0],
                    tag=(0x0062, 0x000B),
                    tag_type="int",
                    default=None
                )

            if current_roi_number is None:
                raise ValueError(
                    f"The current frame ({ii + 1}) has no associated segment number. [{self.describe_self()}]"
                )

            if roi_index == current_roi_number:
                yield ii

    def _get_mask_origin(self, frames: list[int]) -> tuple[float, ...]:
        origin = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0020, 0x0032),
            tag_type="mult_float",
            macro_dcm_seq=(0x0020, 0x9113),
            frame_id=frames[0]
        )[::-1]

        if origin is None:
            raise ValueError(
                f"The DICOM SEG file lacks a information that is required ",
                f"to set the spatial origin of the mask. [{self.describe_self()}]"
            )

        return tuple(origin)

    def _get_mask_spacing(self, frames: list[int]) -> tuple[float, ...]:

        # Get pixel-spacing.
        spacing = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0028, 0x0030),
            tag_type="mult_float",
            macro_dcm_seq=(0x0028, 0x9110),
            frame_id=frames[0]
        )

        if spacing is None:
            raise ValueError(
                f"The DICOM SEG file lacks in-plane mask spacing [{self.describe_self()}]."
            )

        # First try to get spacing between slices.
        z_spacing = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0088),
            tag_type="float",
            macro_dcm_seq=(0x0028, 0x9110),
            frame_id=frames[0]
        )

        # If spacing between slices is not set, get slice thickness.
        if z_spacing is None:
            z_spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0018, 0x0050),
                tag_type="float",
                macro_dcm_seq=(0x0028, 0x9110),
                frame_id=frames[0]
            )

        if z_spacing is None:
            raise ValueError(
                f"The DICOM SEG file lacks slice spacing [{self.describe_self()}]."
            )

        spacing += [z_spacing]

        return tuple(spacing[::-1])

    def _get_mask_orientation(self, frames: list[int]) -> np.array:
        orientation: list[float] = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0020, 0x0037),
            tag_type="mult_float",
            macro_dcm_seq=(0x0020, 0x9116),
            frame_id=frames[0]
        )

        if orientation is None:
            raise ValueError(
                f"The DICOM SEG file lacks information to determine the mask orientation. "
                f"[{self.describe_self()}]"
            )

        # First compute z-orientation.
        # noinspection PyUnreachableCode
        orientation += list(np.cross(orientation[0:3], orientation[3:6]))
        return np.reshape(orientation[::-1], [3, 3], order="F")

    def export_roi_labels(self):

        self.load_metadata()

        # Find which roi numbers (3006,0022) are associated with which roi names (3004,0024).
        labels = [
            get_pydicom_meta_tag(dcm_seq=current_segment_sequence, tag=(0x0062, 0x0005), tag_type="str", default=None)
            for current_segment_sequence in self.image_metadata[(0x0062, 0x0002)]
        ]

        n_labels = max([1, len(labels)])

        if len(labels) == 0:
            labels = [None]

        # Get general attributes.
        parent_attributes = self._get_export_attributes()

        # Add roi labels as attribute.
        attributes = [("roi_label", labels)]
        parent_attributes.update(dict(attributes))

        return parent_attributes



import os.path
import shutil
import sys
import warnings
from pathlib import Path


def get_settings_xml(target_dir: str | Path):
    """
    Creates a local copy of the settings ``xml`` file. This file can be used to configure the image processing and
    feature extraction workflow.

    Parameters
    ----------
    target_dir: str or Path
        Path where the settings ``xml`` file should be copied to.

    Returns
    -------
    None
        No return values. The settings ``xml`` is copied to the intended directory.
    """

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    destination_file_path = os.path.join(target_dir, "settings.xml")
    if os.path.exists(destination_file_path):
        warnings.warn(
            f"A settings xml file already exists at {destination_file_path}.", UserWarning
        )
        return

    # mirp might not be formally installed as a module.
    try:
        mirp_dir = sys.modules["mirp"].__path__[0]
    except KeyError:
        mirp_dir = os.path.dirname(os.path.abspath(os.path.join(__file__, os.path.pardir)))
    source_file_path = os.path.join(mirp_dir, "config_settings.xml")

    shutil.copy(source_file_path, destination_file_path)

    print(f"A copy of the settings xml file was created at {destination_file_path}.")


def get_data_xml(target_dir: str | Path):
    """
    Creates a local copy of the data ``xml`` file. This file can be used to configure import of images and masks.

    Parameters
    ----------
    target_dir: str or Path
        Path where the data ``xml`` file should be copied to.

    Returns
    -------
    None
        No return values. The data ``xml`` is copied to the intended directory.
    """

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    destination_file_path = os.path.join(target_dir, "data.xml")
    if os.path.exists(destination_file_path):
        warnings.warn(
            f"A data xml file already exists at {destination_file_path}.", UserWarning
        )
        return

    # mirp might not be formally installed as a module.
    try:
        mirp_dir = sys.modules["mirp"].__path__[0]
    except KeyError:
        mirp_dir = os.path.dirname(os.path.abspath(os.path.join(__file__, os.path.pardir)))
    source_file_path = os.path.join(mirp_dir, "config_data.xml")

    shutil.copy(source_file_path, destination_file_path)

    print(f"A copy of the data xml file was created at {destination_file_path}.")



import numpy as np

from typing import Any
from dataclasses import dataclass
from mirp.settings.utilities import setting_def


@dataclass
class ImagePostProcessingClass:
    """
    Parameters related to image processing. Note that parameters concerning image perturbation / augmentation and
    resampling are set separately, see :class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass` and
    :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass`.

    Parameters
    ----------
    bias_field_correction: bool, optional, default: False
        Determines whether N4 bias field correction should be performed. When a tissue mask is present, bias field
        correction is conducted using the information contained within the mask.

        .. note::
            Bias-field correction can only be applied to MR imaging.

        .. note::
            Bias-field correction using the N4 algorithm is computationally expensive, and it may be preferable to
            perform this correction prior to feeding MR data to MIRP.

    bias_field_correction_n_fitting_levels: int, optional, default: 3
        The number of fitting levels for the N4 bias field correction algorithm. The default in ITK (1) prevents

    bias_field_correction_n_max_iterations: int or list of int, optional, default: 50
        The number of fitting iterations for the N4 bias field algorithm. A single integer, or a list of integers
        with a length equal to the number of fitting levels is expected.

    bias_field_convergence_threshold: float, optional, default: 0.001
        Convergence threshold for N4 bias field correction algorithm.

    pet_suv_conversion: {"body_weight", "body_surface_area", "lean_body_mass", "lean_body_mass_bmi", "ideal_body_weight". "none"}, default: "body_weight"
        Intensities in PET imaging are often stored as detected radiotracer activity. To make detected activity more
        comparable between patients, these are converted to standardised uptake values. The following are possible:

        * "body_weight": activity is normalised by body weight.
        * "body_surface_area": activity is normalised by body surface area according to DuBois (A formula to estimate
          the approximate surface area if height and weight be known. Arch intern med. 1916;17:863-71).
        * "lean_body_mass": activity is normalised by lean body mass according to James et al. (DHSS/MRC Group on
          Obesity Research, James WP, Waterlow JC. Research on Obesity: A Report of the DHSS/MRC Group; Compiled by
          WPT James. HM Stationery Office; 1976).
        * "lean_body_mass_bmi": activity is normalised by lean body mass according to Janmahasatian et al.
          (Quantification of lean bodyweight. Clinical pharmacokinetics. 2005 Oct;44:1051-65).
        * "ideal_body_weight": activity is normalised by ideal body weight according to Zasadny and Wahl (
          Standardized uptake values of normal tissues at PET with 2-[fluorine-18]-fluoro-2-deoxy-D-glucose:
          variations with body weight and a method for correction. Radiology. 1993 Dec;189(3):847-50).
        * "none": activity is not normalised.

        .. note::
            Conversion of activity to standardised uptake values can only be performed for PET data.

        .. warning::
            Conversion of activity to standardised uptake values requires metadata related to image acquisition and to
            the patient. These metadata are only present in DICOM files. MIRP cannot convert activity to standardised
            uptake values for images in different formats, and this parameter will have no effect.

    intensity_normalisation: {"none", "range", "relative_range", "quantile_range", "standardisation"}, default: "none"
        Specifies the algorithm used to normalise intensities in the image. Will use only intensities in voxels
        masked by the tissue mask (of present). The following are possible:

        * "none": no normalisation
        * "range": normalises intensities based on a fixed mapping against the ``intensity_normalisation_range``
          parameter, which is interpreted to represent an intensity range.
        * "relative_range": normalises intensities based on a fixed mapping against the ``intensity_normalisation_range``
          parameter, which is interpreted to represent a relative intensity range.
        * "quantile_range": normalises intensities based on a fixed mapping against the
          ``intensity_normalisation_range`` parameter, which is interpreted to represent a quantile range.
        * "standardisation": normalises intensities by subtraction of the mean intensity and division by the standard
          deviation of intensities.

        .. note::
            Intensity normalisation may remove any physical meaning of intensity units. For example, intensity
            normalisation of CT images yield intensities that no longer represent Hounsfield Units.

    intensity_normalisation_range: list of float, optional
        Required for "range", "relative_range", and "quantile_range" intensity normalisation methods, and defines the
        intensities that are mapped to the [0.0, 1.0] range during normalisation. The default range depends on the
        type of normalisation method:

        * "range": [np.nan, np.nan]: the minimum and maximum intensity value present in the image are used to set the
          mapping range.
        * "relative_range": [0.0. 1.0]: the minimum (0.0) and maximum (1.0) intensity value present in the image are
          used to set the mapping range.
        * "quantile_range": [0.025, 0.975] the 2.5th and 97.5th percentiles of the intensities in the image are used
          to set the mapping range.

        The lower end of the range is mapped to 0.0 and the upper end to 1.0. However, if intensities below the lower
        end or above the upper end are present in the image, values below 0.0 or above 1.0 may be encountered after
        normalisation. Use ``intensity_normalisation_saturation`` to cap intensities after normalisation to a
        specific range.

    intensity_normalisation_saturation: list of float, optional, default: [np.nan, np.nan]
        Defines the start and endpoint for the saturation range. Normalised intensities that lie outside this
        range are mapped to the limits of the saturation range, e.g. with a range of [0.0, 0.8] all values greater
        than 0.8 are assigned a value of 0.8. np.nan can be used to define limits where the intensity values should
        not be saturated.

    intensity_scaling: float, optional
        Defines scaling parameter to linearly scale intensities with. The scaling parameter is applied after
        normalisation (if any). For example, `intensity_scaling = 1000.0`, combined with `intensity_normalisation =
        "range"` results in intensities being mapped to a [0.0, 1000.0] range instead of [0.0, 1.0].

    tissue_mask_type: {"none", "range", "relative_range"}, optional, default: "relative_range"
        Type of algorithm used to produce an approximate tissue mask of the tissue. Such masks can be used to select
        pixels for bias correction and intensity normalisation by excluding non-tissue voxels.

    tissue_mask_range: list of float, optional
        Range values for creating an approximate mask of the tissue. Required for "range" and "relative_range"
        options. Default: [0.02, 1.00] (``"relative_range"``); [np.nan, np.nan] (``"range"``; effectively all voxels
        are considered to represent tissue).

    **kwargs:
        Unused keyword arguments.
    """

    def __init__(
            self,
            bias_field_correction: bool = False,
            bias_field_correction_n_fitting_levels: int = 3,
            bias_field_correction_n_max_iterations: int | list[int] | None = None,
            bias_field_convergence_threshold: float = 0.001,
            pet_suv_conversion: str = "body_weight",
            intensity_normalisation: str = "none",
            intensity_normalisation_range: list[float] | None = None,
            intensity_normalisation_saturation: list[float] | None = None,
            intensity_scaling: float | None = None,
            tissue_mask_type: str = "relative_range",
            tissue_mask_range: list[float] | None = None,
            **kwargs
    ):

        # Set bias_field_correction parameter
        self.bias_field_correction = bias_field_correction

        # Check n_fitting_levels.
        if bias_field_correction:
            if not isinstance(bias_field_correction_n_fitting_levels, int):
                raise TypeError("The bias_field_correction_n_fitting_levels should be integer with value 1 or larger.")

            if bias_field_correction_n_fitting_levels < 1:
                raise ValueError(
                    f"The bias_field_correction_n_fitting_levels should be integer with value 1 or larger. "
                    f"Found: {bias_field_correction_n_fitting_levels}")

        else:
            bias_field_correction_n_fitting_levels = None

        # Set n_fitting_levels.
        self.n_fitting_levels: None | int = bias_field_correction_n_fitting_levels

        # Set default value for bias_field_correction_n_max_iterations. This is the number of iterations per fitting
        # level.
        if bias_field_correction_n_max_iterations is None and bias_field_correction:
            bias_field_correction_n_max_iterations = [50 for ii in range(bias_field_correction_n_fitting_levels)]

        if bias_field_correction:
            # Parse to list, if a single value is provided.
            if not isinstance(bias_field_correction_n_max_iterations, list):
                bias_field_correction_n_max_iterations = [bias_field_correction_n_max_iterations]

            # Ensure that the list of maximum iteration values equals the number of fitting levels.
            if bias_field_correction_n_fitting_levels > 1 and len(bias_field_correction_n_max_iterations) == 1:
                bias_field_correction_n_max_iterations = [
                    bias_field_correction_n_max_iterations[0]
                    for ii in range(bias_field_correction_n_fitting_levels)
                ]

            # Check that the list of maximum iteration values is equal to the number of fitting levels.
            if len(bias_field_correction_n_max_iterations) != bias_field_correction_n_fitting_levels:
                raise ValueError(
                    f"The bias_field_correction_n_max_iterations parameter should be a list with a length equal to the"
                    f" number of fitting levels ({bias_field_correction_n_fitting_levels}). Found list with "
                    f"{len(bias_field_correction_n_max_iterations)} values.")

            # Check that all values are integers.
            if not all(isinstance(ii, int) for ii in bias_field_correction_n_max_iterations):
                raise TypeError(
                    f"The bias_field_correction_n_max_iterations parameter should be a list of positive "
                    f"integer values. At least one value was not an integer.")

            # Check that all values are positive.
            if not all([ii > 0 for ii in bias_field_correction_n_max_iterations]):
                raise ValueError(
                    f"The bias_field_correction_n_max_iterations parameter should be a list of positive "
                    f"integer values. At least one value was zero or negative.")

        else:
            bias_field_correction_n_max_iterations = None

        # Set n_max_iterations attribute.
        self.n_max_iterations: list[int] | None = bias_field_correction_n_max_iterations

        # Check that the convergence threshold is a non-negative number.
        if bias_field_correction:

            # Check that the value is a float.
            if not isinstance(bias_field_convergence_threshold, float):
                raise TypeError(
                    f"The bias_field_convergence_threshold parameter is expected to be a non-negative "
                    f"floating point value. Found: a value that was not a floating point value.")

            if bias_field_convergence_threshold <= 0.0:
                raise TypeError(
                    f"The bias_field_convergence_threshold parameter is expected to be a non-positive floating point "
                    f"value. Found: a value that was 0.0 or negative ({bias_field_convergence_threshold}).")

        else:
            bias_field_convergence_threshold = None

        # Set convergence_threshold attribute.
        self.convergence_threshold: None | float = bias_field_convergence_threshold

        # Check that pet_suv_conversion has the correct values.
        if pet_suv_conversion not in [
            "body_weight", "body_surface_area", "lean_body_mass", "lean_body_mass_bmi", "ideal_body_weight", "none"
        ]:
            raise ValueError(
                f"The pet_suv_conversion parameter is expected to have one of the following values: ",
                f"'body_weight', `body_surface_area`, `lean_body_mass`, `lean_body_mass_bmi`, `ideal_body_weight` or "
                f"'none'. Found: {pet_suv_conversion}"
            )

        # Set suv_conversion_type parameter.
        self.suv_conversion_type = pet_suv_conversion

        # Check that intensity_normalisation has the correct values.
        if intensity_normalisation not in ["none", "range", "relative_range", "quantile_range", "standardisation"]:
            raise ValueError(
                f"The intensity_normalisation parameter is expected to have one of the following values: "
                f"'none', 'range', 'relative_range', 'quantile_range', 'standardisation'. Found: "
                f"{intensity_normalisation}.")

        # Set intensity_normalisation parameter.
        self.intensity_normalisation = intensity_normalisation

        # Set default value.
        if intensity_normalisation_range is None:
            if intensity_normalisation == "range":
                # Cannot define a proper range.
                intensity_normalisation_range = [np.nan, np.nan]

            elif intensity_normalisation == "relative_range":
                intensity_normalisation_range = [0.0, 1.0]

            elif intensity_normalisation == "quantile_range":
                intensity_normalisation_range = [0.025, 0.975]

        if intensity_normalisation == "range":
            # Check that the range has length 2 and contains floating point values.
            if not isinstance(intensity_normalisation_range, list):
                raise TypeError(
                    f"The intensity_normalisation_range parameter for range-based normalisation should "
                    f"be a list with exactly two values, which are mapped to 0.0 and 1.0 respectively. "
                    f"Found: an object that is not a list.")

            if len(intensity_normalisation_range) != 2:
                raise ValueError(
                    f"The intensity_normalisation_range parameter for range-based normalisation should "
                    f"be a list with exactly two values, which are mapped to 0.0 and 1.0 respectively. "
                    f"Found: list with {len(intensity_normalisation_range)} values.")

            if not all(isinstance(ii, float) for ii in intensity_normalisation_range):
                raise TypeError(
                    f"The intensity_normalisation_range parameter for range-based normalisation should "
                    f"be a list with exactly two floating point values, which are mapped to 0.0 and 1.0 "
                    f"respectively. Found: one or more values that are not floating point values.")

        elif intensity_normalisation in ["relative_range", "quantile_range"]:
            # Check that the range has length 2 and contains floating point values between 0.0 and 1.0.
            if intensity_normalisation == "relative_range":
                intensity_normalisation_specifier = "relative range-based normalisation"
            else:
                intensity_normalisation_specifier = "quantile range-based normalisation"

            if not isinstance(intensity_normalisation_range, list):
                raise TypeError(
                    f"The intensity_normalisation_range parameter for {intensity_normalisation_specifier} "
                    f"should be a list with exactly two values, which are mapped to 0.0 and 1.0 "
                    f"respectively. Found: an object that is not a list.")

            if len(intensity_normalisation_range) != 2:
                raise ValueError(
                    f"The intensity_normalisation_range parameter for {intensity_normalisation_specifier} "
                    f"should be a list with exactly two values, which are mapped to 0.0 and 1.0 "
                    f"respectively. Found: list with {len(intensity_normalisation_range)} values.")

            if not all(isinstance(ii, float) for ii in intensity_normalisation_range):
                raise TypeError(
                    f"The intensity_normalisation_range parameter for {intensity_normalisation_specifier} "
                    f"should be a list with exactly two values, which are mapped to 0.0 and 1.0 "
                    f"respectively. Found: one or more values that are not floating point values.")

            if not all([0.0 <= ii <= 1.0 for ii in intensity_normalisation_range]):
                raise TypeError(
                    f"The intensity_normalisation_range parameter for {intensity_normalisation_specifier} "
                    f"should be a list with exactly two values, which are mapped to 0.0 and 1.0 "
                    f"respectively. Found: one or more values that are outside the [0.0, 1.0] range.")

        else:
            # None and standardisation do not use this range.
            intensity_normalisation_range = None

        # Set normalisation range.
        self.intensity_normalisation_range: None | list[float] = intensity_normalisation_range

        # Check intensity normalisation saturation range.
        if intensity_normalisation_saturation is None:
            intensity_normalisation_saturation = [np.nan, np.nan]

        if not isinstance(intensity_normalisation_saturation, list):
            raise TypeError(
                "The intensity_normalisation_saturation parameter is expected to be a "
                "list of two floating point values."
            )

        if not len(intensity_normalisation_saturation) == 2:
            raise ValueError(
                f"The intensity_normalisation_saturation parameter should consist of two values. Found: "
                f"{len(intensity_normalisation_saturation)} values.")

        if not all(isinstance(ii, float) for ii in intensity_normalisation_saturation):
            raise TypeError(
                "The intensity_normalisation_saturation parameter can only contain floating point or np.nan values."
            )

        # intensity_normalisation_saturation parameter
        self.intensity_normalisation_saturation: None | list[float] = intensity_normalisation_saturation

        # Check intensity_scaling
        if intensity_scaling is not None:
            if not isinstance(intensity_scaling, float):
                raise TypeError("The intensity_scaling parameter is expected to be a single floating point.")
            if intensity_scaling == 0.0:
                raise ValueError("The intensity_scaling parameter cannot have a value of 0.0.")
        else:
            intensity_scaling = 1.0

        self.intensity_scaling: float = intensity_scaling

        # Check tissue_mask_type
        if tissue_mask_type not in ["none", "range", "relative_range"]:
            raise ValueError(
                f"The tissue_mask_type parameter is expected to have one of the following values: "
                f"'none', 'range', or 'relative_range'. Found: {tissue_mask_type}."
            )

        # Set tissue_mask_type
        self.tissue_mask_type: str = tissue_mask_type

        # Set the default value for tissue_mask_range.
        if tissue_mask_range is None:
            if tissue_mask_type == "relative_range":
                tissue_mask_range = [0.02, 1.00]
            elif tissue_mask_type == "range":
                tissue_mask_range = [np.nan, np.nan]
            else:
                tissue_mask_range = [np.nan, np.nan]

        # Perform checks on tissue_mask_range.
        if tissue_mask_type != "none":
            if not isinstance(tissue_mask_range, list):
                raise TypeError(
                    "The tissue_mask_range parameter is expected to be a list of two floating point values.")

            if not len(tissue_mask_range) == 2:
                raise ValueError(
                    f"The tissue_mask_range parameter should consist of two values. Found: "
                    f"{len(tissue_mask_range)} values.")

            if not all(isinstance(ii, float) for ii in tissue_mask_range):
                raise TypeError("The tissue_mask_range parameter can only contain floating point or np.nan values.")

            if tissue_mask_type == "relative_range":
                if not all([(0.0 <= ii <= 1.0) or np.isnan(ii) for ii in tissue_mask_range]):
                    raise ValueError(
                        "The tissue_mask_range parameter should consist of two values between 0.0 and 1.0.")

        # Set tissue_mask_range.
        self.tissue_mask_range: tuple[float, ...] = tuple(tissue_mask_range)


def get_post_processing_settings() -> list[dict[str, Any]]:

    return [
        setting_def("bias_field_correction", "bool", test=True),
        setting_def(
            "bias_field_correction_n_fitting_levels", "int", xml_key="n_fitting_levels",
            class_key="n_fitting_levels", test=2
        ),
        setting_def(
            "bias_field_correction_n_max_iterations", "int", xml_key="n_max_iterations",
            class_key="n_max_iterations", to_list=True, test=[1000, 1000]
        ),
        setting_def(
            "bias_field_convergence_threshold", "float", xml_key="convergence_threshold",
            class_key="convergence_threshold", test=0.1
        ),
        setting_def("pet_suv_conversion", "str", class_key="suv_conversion_type", test="none"),
        setting_def("intensity_normalisation", "str", test="relative_range"),
        setting_def("intensity_normalisation_range", "float", to_list=True, test=[0.10, 0.90]),
        setting_def("intensity_normalisation_saturation", "float", to_list=True, test=[0.00, 10.00]),
        setting_def("intensity_scaling", "float", test=3.0),
        setting_def("tissue_mask_type", "str", test="range"),
        setting_def("tissue_mask_range", "float", to_list=True, test=[0.00, 10.00])
    ]



import numpy as np
from typing import Any
from mirp._images.generic_image import GenericImage


class RTDoseImage(GenericImage):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def get_default_lowest_intensity():
        return 0.0

    @staticmethod
    def get_default_ivh_discretisation_method():
        return "fixed_bin_size"

    @staticmethod
    def get_default_ivh_bin_size():
        return 0.20

    def normalise_intensities(
            self,
            normalisation_method: None | str = "none",
            intensity_range: None | tuple[Any, Any] = None,
            saturation_range: None | tuple[Any, Any] = None,
            mask: None | np.ndarray = None
    ):
        """
        Normalise intensities. NOTE: this changes the class of the object from RTDoseImage to GenericImage as
        normalisation breaks the one-to-one relationship between intensities and dose units.
        """
        if self.image_data is None:
            return self

        if normalisation_method is None or normalisation_method == "none":
            return self

        new_image = GenericImage(image_data=self.image_data)
        new_image.update_from_template(template=self)
        new_image = new_image.normalise_intensities(
            normalisation_method=normalisation_method,
            intensity_range=intensity_range,
            saturation_range=saturation_range,
            mask=mask
        )

        return new_image

    def scale_intensities(self, scale: float):
        if self.image_data is None:
            return self

        if scale == 1.0:
            return self

        # Scaling intensities changes the object class from RTDoseImage to GenericImage as this breaks the one-to-one
        # relationship between intensities and dose.
        new_image = GenericImage(image_data=self.image_data)
        new_image.update_from_template(template=self)
        new_image = new_image.scale_intensities(scale=scale)

        return new_image



import numpy as np
import copy

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import GaussianTransformedImage
from mirp._imagefilters.generic import GenericFilter
from mirp._imagefilters.utilities import FilterSet2D, FilterSet3D
from mirp.settings.generic import SettingsClass


class GaussianFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):
        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = True
        self.ibsi_id = "8BC3"

        self.sigma = settings.img_transform.gaussian_sigma
        self.sigma_cutoff = settings.img_transform.gaussian_sigma_truncate
        self.mode = settings.img_transform.gaussian_boundary_condition

        # Riesz transformation settings.
        self.riesz_order: None | list[int] | list[list[int]] = None
        self.riesz_steered: bool = False
        self.riesz_sigma: None | float | list[float] = None
        if settings.img_transform.has_riesz_filter(x=name):
            self.riesz_order = settings.img_transform.riesz_order

            if settings.img_transform.has_steered_riesz_filter(x=name):
                self.riesz_steered = True
                self.riesz_sigma = settings.img_transform.riesz_filter_tensor_sigma

            # Riesz transformed filters are not IBSI-compliant
            self.ibsi_compliant = False

    def generate_object(self):
        # Generator for transformation objects.
        sigma = copy.deepcopy(self.sigma)
        if not isinstance(sigma, list):
            sigma = [sigma]

        riesz_order = copy.deepcopy(self.riesz_order)
        if riesz_order is None:
            riesz_order = [None]
        elif not all(isinstance(riesz_order_set, list) for riesz_order_set in riesz_order):
            riesz_order = [riesz_order]

        riesz_sigma = copy.deepcopy(self.riesz_sigma)
        if not isinstance(riesz_sigma, list):
            riesz_sigma = [riesz_sigma]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_sigma in sigma:
            for current_riesz_order in riesz_order:
                for current_riesz_sigma in riesz_sigma:
                    filter_object = copy.deepcopy(self)
                    filter_object.sigma = current_sigma
                    filter_object.riesz_order = current_riesz_order
                    filter_object.riesz_sigma = current_riesz_sigma

                    yield filter_object

    def transform(self, image: GenericImage) -> GaussianTransformedImage:
        # Create placeholder Gaussian response map.
        response_map = GaussianTransformedImage(
            image_data=None,
            sigma_parameter=self.sigma,
            sigma_cutoff_parameter=self.sigma_cutoff,
            boundary_condition=self.mode,
            riesz_order=self.riesz_order,
            riesz_steering=self.riesz_steered,
            riesz_sigma_parameter=self.riesz_sigma,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Calculate sigma for current image
        voxel_sigma = np.divide(
            np.full(shape=3, fill_value=self.sigma),
            np.array(image.image_spacing))

        # Apply filters
        response_map.set_voxel_grid(voxel_grid=self.transform_grid(
            voxel_grid=image.get_voxel_grid(),
            sigma=voxel_sigma)
        )

        return response_map

    def transform_grid(self,
                       voxel_grid: np.array,
                       sigma: np.array):
        # Determine the size of the filter
        filter_size = 1 + 2 * np.floor(self.sigma_cutoff * sigma + 0.5)
        filter_size.astype(int)

        if self.separate_slices:
            # Set the number of dimensions.
            d = 2.0

            # Create the grid coordinates, with [0, 0, 0] in the center.
            y, x = np.mgrid[:filter_size[1], :filter_size[2]]
            y -= (filter_size[1] - 1.0) / 2.0
            x -= (filter_size[2] - 1.0) / 2.0

            # Compute the square of the norm.
            norm_2 = np.power(y, 2.0) + np.power(x, 2.0)

        else:
            # Set the number of dimensions.
            d = 3.0

            # Create the grid coordinates, with [0, 0, 0] in the center.
            z, y, x = np.mgrid[:filter_size[0], :filter_size[1], :filter_size[2]]
            z -= (filter_size[0] - 1.0) / 2.0
            y -= (filter_size[1] - 1.0) / 2.0
            x -= (filter_size[2] - 1.0) / 2.0

            # Compute the square of the norm.
            norm_2 = np.power(z, 2.0) + np.power(y, 2.0) + np.power(x, 2.0)

        # Set a single sigma value.
        sigma = np.max(sigma)

        # Compute the scale factor
        scale_factor = np.power(1.0 / np.sqrt(2.0 * np.pi * sigma ** 2), d)

        # Compute the exponent which determines filter width.
        width_factor = - norm_2 / (2.0 * sigma ** 2.0)

        # Compute the weights of the filter.
        filter_weights = np.multiply(scale_factor, np.exp(width_factor))

        if self.separate_slices:
            # Set filter weights and create a filter.
            gauss_filter = FilterSet2D(filter_weights,
                                       riesz_order=self.riesz_order,
                                       riesz_steered=self.riesz_steered,
                                       riesz_sigma=self.riesz_sigma)

            # Convolve laplacian of gaussian filter with the image.
            response_map = gauss_filter.convolve(voxel_grid=voxel_grid,
                                                 mode=self.mode,
                                                 response="real")

        else:
            # Set filter weights and create a filter.
            gauss_filter = FilterSet3D(filter_weights,
                                       riesz_order=self.riesz_order,
                                       riesz_steered=self.riesz_steered,
                                       riesz_sigma=self.riesz_sigma)

            # Convolve laplacian of gaussian filter with the image.
            response_map = gauss_filter.convolve(voxel_grid=voxel_grid,
                                                 mode=self.mode,
                                                 response="real")

        # Compute the convolution
        return response_map



from functools import lru_cache
from typing import Generator

import numpy as np
import pandas as pd


from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import HistogramDerivedFeature
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class DataIntensityVolumeHistogram(object):

    def __init__(self):
        # Raw intensity data
        self.data: pd.DataFrame | None = None

        # Number of voxels
        self.n_voxels: int | None = None

        # Maximum intensity
        self.next_max_intensity: float | int | None = None

    def compute(
            self,
            image: GenericImage,
            mask: BaseMask,
            discretisation_method: str,
            bin_number: int | None = None,
            bin_size: float | None = None
    ):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Convert image volume to table.
        data = mask.as_pandas_dataframe(image=image, intensity_mask=True)
        data = data[data.roi_int_mask == True]

        # Set number of voxels.
        self.n_voxels = len(data)

        # Set intensity range
        if mask.intensity_range is None:
            intensity_range = np.array([np.nan, np.nan])
        else:
            intensity_range = np.array(mask.intensity_range).astype(float)

        if np.isnan(intensity_range[0]):
            intensity_range[0] = float(np.min(data.g))
        if np.isnan(intensity_range[1]):
            intensity_range[1] = float(np.max(data.g))

        # Determine the discretisation method.
        if discretisation_method == "none":
            discretisation_method = image.get_default_ivh_discretisation_method()

        # Set bin_number, if required.
        if discretisation_method == "fixed_bin_number" and bin_number is None:
            bin_number = image.get_default_ivh_bin_number()
            if bin_number is None:
                raise ValueError("IVH bin number should be provided.")

        # Set bin_size, if required.
        if discretisation_method == "fixed_bin_size" and bin_size is None:
            bin_size = image.get_default_ivh_bin_size()
            if bin_size is None:
                raise ValueError("IVH bin size should be provided.")

        if discretisation_method == "none":
            # No transformation.

            # Set number of bins. The number of bins is equal to the number of grey levels present.
            bin_number = intensity_range[1] - intensity_range[0] + 1.0

            # Create histogram by grouping by intensity level and counting bin size
            data = data.groupby(by="g").size().reset_index(name="n")

            # Append empty grey levels to histogram
            levels = np.arange(start=intensity_range[0], stop=intensity_range[1] + 1)
            missing_levels = levels[np.logical_not(np.isin(levels, data.g))]
            n_missing = len(missing_levels)
            if n_missing > 0:
                data = pd.concat([
                    data,
                    pd.DataFrame({"g": missing_levels, "n": np.zeros(n_missing)})
                ], ignore_index=True)

            # Set maximum intensity.
            self.next_max_intensity = intensity_range[1] + 1.0

        elif discretisation_method == "fixed_bin_size":
            # Fixed bin size/width calculations

            # Get the number of bins
            bin_number = np.ceil((intensity_range[1] - intensity_range[0]) / bin_size) + 1.0

            # Bin voxels
            data.g = np.floor((data.g - intensity_range[0]) / (bin_size * 1.0)) + 1.0

            # Set voxels with grey level lower than 0.0 to 1.0. This may occur with non-mask voxels
            # and voxels with the minimum intensity
            data.loc[data["g"] <= 0.0, "g"] = 1.0

            # Create histogram by grouping by intensity level and counting bin size
            data = data.groupby(by="g").size().reset_index(name="n")

            # Append empty grey levels to histogram
            levels = np.arange(start=1, stop=int(bin_number) + 1)
            missing_levels = levels[np.logical_not(np.isin(levels, data.g))]
            n_missing = len(missing_levels)
            if n_missing > 0:
                data = pd.concat([
                    data,
                    pd.DataFrame({"g": missing_levels, "n": np.zeros(n_missing)})
                ], ignore_index=True)

            # Replace g by the bin centers
            data.loc[:, "g"] = intensity_range[0] + (data["g"] - 0.5) * bin_size * 1.0

            # Update intensity range
            intensity_range[0] = np.min(data.g)
            intensity_range[1] = np.max(data.g)

            # Set maximum intensity.
            self.next_max_intensity = intensity_range[1] + bin_size

        elif discretisation_method == "fixed_bin_number":
            # Calculation for all other image types

            # For FBN, the intensity range is always determined by the intensity range in the image.
            intensity_range[0] = np.min(data.g)
            intensity_range[1] = np.max(data.g)

            data.loc[:, "g"] = np.floor(
                bin_number * (data["g"] - intensity_range[0]) / (intensity_range[1] - intensity_range[0])) + 1.0

            # Update values at the boundaries
            data.loc[data["g"] <= 0.0, "g"] = 1.0
            data.loc[data["g"] >= bin_number * 1.0, "g"] = bin_number * 1.0

            # Create histogram by grouping by intensity level and counting bin size
            data = data.groupby(by="g").size().reset_index(name="n")

            # Append empty grey levels to histogram
            levels = np.arange(start=1, stop=bin_number + 1)
            missing_levels = levels[np.logical_not(np.isin(levels, data.g))]
            n_missing = len(missing_levels)
            if n_missing > 0:
                data = pd.concat([
                    data,
                    pd.DataFrame({"g": missing_levels, "n": np.zeros(n_missing)})
                ], ignore_index=True)

            # Update grey level range
            intensity_range[0] = 1.0
            intensity_range[1] = bin_number

            # Set maximum intensity.
            self.next_max_intensity = bin_number + 1.0

        else:
            raise ValueError(f"{discretisation_method} is not a valid IVH discretisation method.")

        # Order histogram table by increasing grey level
        data = data.sort_values(by="g")

        # Calculate intensity fraction
        data["gamma"] = (data.g - intensity_range[0]) / (intensity_range[1] - intensity_range[0])

        # Calculate volume fraction
        data["nu"] = 1.0 - (np.cumsum(np.append([0], data.n))[0:int(bin_number)]) * 1.0 / (self.n_voxels * 1.0)

        # Set data.
        self.data = data

    def is_empty(self):
        return self.data is None or len(self.data) == 0


class FeatureIntensityVolumeHistogram(HistogramDerivedFeature):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask,
            discretisation_method: str,
            bin_number: int | None = None,
            bin_size: float | None = None
    ) -> DataIntensityVolumeHistogram:
        data = DataIntensityVolumeHistogram()
        data.compute(
            image=image,
            mask=mask,
            discretisation_method=discretisation_method,
            bin_number=bin_number,
            bin_size=bin_size
        )

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureIntensityVolumeHistogram):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_number=self.bin_number,
            bin_size=self.bin_width
        )

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = self._compute(data=data)

    def _compute(self, data: DataIntensityVolumeHistogram):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = self._get_base_table_name_element()
        self.table_name = "_".join(table_elements)


class FeatureIVHVolumeFractionAtIntensity(FeatureIntensityVolumeHistogram):
    def __init__(self, percentile: float, **kwargs):
        # This class allows for passing percentile values.
        super().__init__(**kwargs)

        # Set percentile
        self.percentile = percentile

        # Set IBSI identifier. Only the 10th and 90th percentile are explicitly identified.
        ibsi_id = ""
        if percentile == 10.0:
            ibsi_id = "NK6P"
        elif percentile == 90.0:
            ibsi_id = "4279"
        self.ibsi_id = ibsi_id

        if percentile.is_integer():
            self.name = f"IVH - volume fraction at {int(percentile)}% intensity"
            self.abbr_name = f"ivh_v{int(percentile)}"
        else:
            self.name = f"IVH - volume fraction at {percentile}% intensity"
            self.abbr_name = f"ivh_v{percentile}"

        self.ibsi_compliant = True

    def _compute(self, data: DataIntensityVolumeHistogram):
        return data.data.loc[data.data.gamma >= self.percentile / 100.0, :].nu.max()


class FeatureIVHIntensityAtVolumeFraction(FeatureIntensityVolumeHistogram):
    def __init__(self, percentile: float, **kwargs):
        # This class allows for passing percentile values.
        super().__init__(**kwargs)

        # Set percentile
        self.percentile = percentile

        # Set IBSI identifier. Only the 10th and 90th percentile are explicitly identified.
        ibsi_id = ""
        if percentile == 10.0:
            ibsi_id = "PWN1"
        elif percentile == 90.0:
            ibsi_id = "BOHI"
        self.ibsi_id = ibsi_id

        if percentile.is_integer():
            self.name = f"IVH - intensity at {int(percentile)}% volume"
            self.abbr_name = f"ivh_i{int(percentile)}"
        else:
            self.name = f"IVH - intensity at {percentile}% volume"
            self.abbr_name = f"ivh_i{percentile}"

        self.ibsi_compliant = True

    def _compute(self, data: DataIntensityVolumeHistogram):
        x = data.data.loc[data.data.nu <= self.percentile / 100.0, :].g.min()
        if np.isnan(x):
            x = data.next_max_intensity

        return x


class FeatureIVHVolumeFractionDifference(FeatureIntensityVolumeHistogram):
    def __init__(self, percentile: tuple[float, float], **kwargs):
        # This class allows for passing percentile values.
        super().__init__(**kwargs)

        # Set percentile
        self.percentile = percentile

        # Set IBSI identifier. Only the 10th and 90th percentile are explicitly identified.
        ibsi_id = ""
        if percentile[0] == 10.0 and percentile[1] == 90.0:
            ibsi_id = "WITY"
        self.ibsi_id = ibsi_id

        if all(x.is_integer() for x in percentile):
            self.name = f"IVH - difference in volume fraction between {int(percentile[0])}% and {int(percentile[1])}% intensity"
            self.abbr_name = f"ivh_diff_v{int(percentile[0])}_v{int(percentile[1])}"
        else:
            self.name = f"IVH - difference in volume fraction between {percentile[0]}% and {percentile[1]}% intensity"
            self.abbr_name = f"ivh_diff_v{percentile[0]}_v{percentile[1]}"

        self.ibsi_compliant = True

    def _compute(self, data: DataIntensityVolumeHistogram):
        return (
                data.data.loc[data.data.gamma >= self.percentile[0] / 100.0, :].nu.max()
                - data.data.loc[data.data.gamma >= self.percentile[1] / 100.0, :].nu.max()
        )


class FeatureIVHIntensityDifference(FeatureIntensityVolumeHistogram):
    def __init__(self, percentile: tuple[float, float], **kwargs):
        # This class allows for passing percentile values.
        super().__init__(**kwargs)

        # Set percentile
        self.percentile = percentile

        # Set IBSI identifier. Only the 10th and 90th percentile are explicitly identified.
        ibsi_id = ""
        if percentile[0] == 10.0 and percentile[1] == 90.0:
            ibsi_id = "JXJA"
        self.ibsi_id = ibsi_id

        if all(x.is_integer() for x in percentile):
            self.name = f"IVH - difference in intensity between {int(percentile[0])}% and {int(percentile[1])}% volume"
            self.abbr_name = f"ivh_diff_i{int(percentile[0])}_i{int(percentile[1])}"
        else:
            self.name = f"IVH - difference in intensity between {percentile[0]}% and {percentile[1]}% volume"
            self.abbr_name = f"ivh_diff_i{percentile[0]}_i{percentile[1]}"

        self.ibsi_compliant = True

    def _compute(self, data: DataIntensityVolumeHistogram):
        x_0 = data.data.loc[data.data.nu <= self.percentile[0] / 100.0, :].g.min()
        if np.isnan(x_0):
            x_0 = data.next_max_intensity

        x_1 = data.data.loc[data.data.nu <= self.percentile[1] / 100.0, :].g.min()
        if np.isnan(x_1):
            x_1 = data.next_max_intensity

        return x_0 - x_1


class FeatureIVHAreaUnderCurve(FeatureIntensityVolumeHistogram):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IVH - area under IVH curve"
        self.abbr_name = "ivh_auc"
        self.ibsi_id = "9CMM"
        self.ibsi_compliant = False

    def _compute(self, data: DataIntensityVolumeHistogram):
        return np.trapezoid(y=data.data.nu, x=data.data.gamma)


def get_intensity_volume_histogram_class_dict() -> dict[str, FeatureIntensityVolumeHistogram]:
    class_dict = {
        "ivh_v": FeatureIVHVolumeFractionAtIntensity,
        "ivh_i": FeatureIVHIntensityAtVolumeFraction,
        "ivh_diff_v": FeatureIVHVolumeFractionDifference,
        "ivh_diff_i": FeatureIVHIntensityDifference,
        "ivh_auc": FeatureIVHAreaUnderCurve
    }

    return class_dict


def generate_ivh_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureIntensityVolumeHistogram, None, None]:
    class_dict = get_intensity_volume_histogram_class_dict()
    ivh_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_ivh_family():
        features = ivh_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only ivh features, and return if none are present.
    features = [feature for feature in features if feature in ivh_features]
    if len(features) == 0:
        return

    # Set default percentiles.
    intensity_percentiles = [10.0, 25.0, 50.0, 75.0, 90.0]
    volume_percentiles = [10.0, 25.0, 50.0, 75.0, 90.0]
    intensity_difference_range = [(10.0, 90.0), (25.0, 75.0)]
    volume_difference_range = [(10.0, 90.0), (25.0, 75.0)]

    discretisation_parameters = {
        "discretisation_method": settings.ivh_discretisation_method,
        "bin_width": settings.ivh_discretisation_bin_width,
        "bin_number": settings.ivh_discretisation_n_bins
    }

    for feature in features:
        if feature == "ivh_v":
            for percentile in volume_percentiles:
                yield class_dict[feature](
                    percentile=percentile,
                    **discretisation_parameters
                )
        elif feature == "ivh_i":
            for percentile in intensity_percentiles:
                yield class_dict[feature](
                    percentile=percentile,
                    **discretisation_parameters
                )
        elif feature == "ivh_diff_v":
            for percentile in volume_difference_range:
                yield class_dict[feature](
                    percentile=percentile,
                    **discretisation_parameters
                )
        elif feature == "ivh_diff_i":
            for percentile in intensity_difference_range:
                yield class_dict[feature](
                    percentile=percentile,
                    **discretisation_parameters
                )
        else:
            yield class_dict[feature](**discretisation_parameters)



from mirp._data_import.generic_file import ImageFile


class BaseWorkflow:
    def __init__(
            self,
            image_file: ImageFile,
            write_dir: None | str = None,
            **kwargs
    ):
        super().__init__()
        self.image_file = image_file
        self.write_dir = write_dir



import copy
import pandas as pd
from typing import Any

from mirp._images.generic_image import GenericImage


class TransformedImage(GenericImage):
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    def get_default_ivh_discretisation_method():
        return "fixed_bin_number"


class GaborTransformedImage(TransformedImage):
    def __init__(
            self,
            sigma_parameter: None | float = None,
            gamma_parameter: None | float = None,
            lambda_parameter: None | float = None,
            theta_parameter: None | float = None,
            pool_theta: None | bool = None,
            response_type: None | str = None,
            rotation_invariance: None | bool = None,
            pooling_method: None | str = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.sigma_parameter = sigma_parameter
        self.gamma_parameter = gamma_parameter
        self.lambda_parameter = lambda_parameter
        self.theta_parameter = theta_parameter
        self.pool_theta = pool_theta
        self.response_type = response_type
        self.rotation_invariance = rotation_invariance
        self.pooling_method = pooling_method
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        descriptors += [
            "gabor",
            "s", str(self.sigma_parameter),
            "g", str(self.gamma_parameter),
            "l", str(self.lambda_parameter)
        ]

        if not self.pool_theta:
            descriptors += ["t", str(self.theta_parameter)]

        descriptors += ["2D" if self.separate_slices else "3D"]

        if self.rotation_invariance and not self.separate_slices:
            descriptors += ["invar"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "gabor"),
            ("sigma_parameter", self.sigma_parameter),
            ("gamma_parameter", self.gamma_parameter),
            ("lambda_parameter", self.lambda_parameter),
            ("theta_parameter", self.theta_parameter),
            ("pool_theta", self.pool_theta),
            ("response_type", self.response_type),
            ("rotation_invariance", self.rotation_invariance),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.pooling_method is not None:
            attributes += [("pooling_method", self.pooling_method)]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = [
            "gabor",
            "s", str(self.sigma_parameter),
            "g", str(self.gamma_parameter),
            "l", str(self.lambda_parameter)
        ]

        if not self.pool_theta:
            feature_name_prefix += ["t", str(self.theta_parameter)]

        feature_name_prefix += ["2D" if self.separate_slices else "3D"]

        if self.rotation_invariance and not self.separate_slices:
            feature_name_prefix += ["invar"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class GaussianTransformedImage(TransformedImage):
    def __init__(
            self,
            sigma_parameter: None | float = None,
            sigma_cutoff_parameter: None | float = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.sigma_parameter = sigma_parameter
        self.sigma_cutoff_parameter = sigma_cutoff_parameter
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        descriptors += [
            "gaussian",
            "s", str(self.sigma_parameter)
        ]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "gaussian"),
            ("sigma_parameter", self.sigma_parameter),
            ("sigma_cutoff_parameter", self.sigma_cutoff_parameter),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = [
            "gaussian",
            "s", str(self.sigma_parameter)
        ]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class LaplacianOfGaussianTransformedImage(TransformedImage):
    def __init__(
            self,
            sigma_parameter: None | float = None,
            sigma_cutoff_parameter: None | float = None,
            pooling_method: None | str = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.sigma_parameter = sigma_parameter
        self.sigma_cutoff_parameter = sigma_cutoff_parameter
        self.pooling_method = pooling_method
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        descriptors += [
            "log",
            "s", str(self.sigma_parameter)
        ]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "laplacian_of_gaussian"),
            ("sigma_parameter", self.sigma_parameter),
            ("sigma_cutoff_parameter", self.sigma_cutoff_parameter),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.pooling_method is not None:
            attributes += [("pooling_method", self.pooling_method)]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = [
            "log",
            "s", str(self.sigma_parameter)
        ]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class LawsTransformedImage(TransformedImage):
    def __init__(
            self,
            laws_kernel: None | str | list[str] = None,
            delta_parameter: None | int = None,
            energy_map: None | bool = None,
            rotation_invariance: None | bool = None,
            pooling_method: None | str = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.laws_kernel = laws_kernel
        self.delta_parameter = delta_parameter
        self.energy_map = energy_map
        self.rotation_invariance = rotation_invariance
        self.pooling_method = pooling_method
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        descriptors += ["laws", self.laws_kernel]
        if self.energy_map:
            descriptors += ["energy", "delta", str(self.delta_parameter)]
        if self.rotation_invariance:
            descriptors += ["invar"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "laws"),
            ("laws_kernel", self.laws_kernel),
            ("energy_map", self.energy_map),
            ("rotation_invariance", self.rotation_invariance),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.energy_map:
            attributes += [("delta_parameter", self.delta_parameter)]

        if self.pooling_method is not None:
            attributes += [("pooling_method", self.pooling_method)]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = ["laws", self.laws_kernel]
        if self.energy_map:
            feature_name_prefix += ["energy", "delta", str(self.delta_parameter)]
        if self.rotation_invariance:
            feature_name_prefix += ["invar"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class MeanTransformedImage(TransformedImage):
    def __init__(
            self,
            filter_size: None | int = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.filter_size = filter_size
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()
        descriptors += ["mean", "d", str(self.filter_size)]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "mean"),
            ("filter_size", self.filter_size),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = ["mean", "d", str(self.filter_size)]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class NonSeparableWaveletTransformedImage(TransformedImage):
    def __init__(
            self,
            wavelet_family: None | str = None,
            decomposition_level: None | int = None,
            response_type: None | str = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.wavelet_family = wavelet_family
        self.decomposition_level = decomposition_level
        self.response_type = response_type
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        descriptors += [
            "wavelet", self.wavelet_family,
            "level", str(self.decomposition_level)
        ]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "non_separable_wavelet"),
            ("wavelet_family", self.wavelet_family),
            ("decomposition_level", self.decomposition_level),
            ("response_type", self.response_type),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = [
            "wavelet", self.wavelet_family,
            "level", str(self.decomposition_level)
        ]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class SeparableWaveletTransformedImage(TransformedImage):
    def __init__(
            self,
            wavelet_family: None | str = None,
            decomposition_level: None | int = None,
            filter_kernel_set: None | str = None,
            stationary_wavelet: None | bool = None,
            rotation_invariance: None | bool = None,
            pooling_method: None | str = None,
            boundary_condition: None | str = None,
            riesz_order: None | int | list[int] = None,
            riesz_steering: None | bool = None,
            riesz_sigma_parameter: None | float = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Filter parameters
        self.wavelet_family = wavelet_family
        self.decomposition_level = decomposition_level
        self.filter_kernel_set = filter_kernel_set
        self.stationary_wavelet = stationary_wavelet
        self.rotation_invariance = rotation_invariance
        self.pooling_method = pooling_method
        self.boundary_condition = boundary_condition
        self.riesz_transformed = riesz_order is not None
        self.riesz_order = copy.deepcopy(riesz_order)
        self.riesz_steering = riesz_steering
        self.riesz_sigma_parameter = riesz_sigma_parameter

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()

        descriptors += [
            "wavelet", self.wavelet_family, self.filter_kernel_set,
            "level", str(self.decomposition_level)
        ]

        if not self.stationary_wavelet:
            descriptors += ["decimated"]
        if self.rotation_invariance:
            descriptors += ["invar"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()

        attributes = [
            ("filter_type", "separable_wavelet"),
            ("wavelet_family", self.wavelet_family),
            ("wavelet_kernel", self.filter_kernel_set),
            ("decomposition_level", self.decomposition_level),
            ("stationary_wavelet", self.stationary_wavelet),
            ("rotation_invariance", self.rotation_invariance),
            ("boundary_condition", self.boundary_condition)
        ]

        if self.pooling_method is not None:
            attributes += [("pooling_method", self.pooling_method)]

        if self.riesz_transformed:
            attributes += [("riesz_order", self.riesz_order)]

            if self.riesz_steering:
                attributes += [("riesz_sigma_parameter", self.riesz_sigma_parameter)]

        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)

        feature_name_prefix = [
            "wavelet", self.wavelet_family, self.filter_kernel_set,
            "level", str(self.decomposition_level)
        ]

        if not self.stationary_wavelet:
            feature_name_prefix += ["decimated"]
        if self.rotation_invariance:
            feature_name_prefix += ["invar"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class SquareTransformedImage(TransformedImage):
    def __init__(
            self,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()
        descriptors += ["square"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()
        attributes = [("filter_type", "square_transformation")]
        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)
        feature_name_prefix = ["square"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class SquareRootTransformedImage(TransformedImage):
    def __init__(
            self,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()
        descriptors += ["sqrt"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()
        attributes = [("filter_type", "square_root_transformation")]
        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)
        feature_name_prefix = ["sqrt"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class LogarithmTransformedImage(TransformedImage):
    def __init__(
            self,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()
        descriptors += ["lgrthm"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()
        attributes = [("filter_type", "logarithm_transformation")]
        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)
        feature_name_prefix = ["lgrthm"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class ExponentialTransformedImage(TransformedImage):
    def __init__(
            self,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()
        descriptors += ["exp"]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()
        attributes = [("filter_type", "exponential_transformation")]
        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)
        feature_name_prefix = ["exp"]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x


class LocalBinaryPatternImage(TransformedImage):
    def __init__(
            self,
            distance: None | float = None,
            separate_slices: None | bool = None,
            lbp_method: None | str = None,
            template: None | GenericImage = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Update image parameters using the template.
        if isinstance(template, GenericImage):
            self.update_from_template(template=template)

        self.distance = distance
        self.separate_slices = separate_slices
        self.lbp_method = lbp_method

    def get_file_name_descriptor(self) -> list[str]:
        descriptors = super().get_file_name_descriptor()
        descriptors += [
            "lbp",
            "2d" if self.separate_slices else "3d"
        ]

        # Don't add anything if the method is "default".
        if self.lbp_method == "variance":
            descriptors += ["var"]
        elif self.lbp_method == "rotation_invariant":
            descriptors += ["rot_invar"]

        descriptors += ["d" + str(self.distance)]

        return descriptors

    def get_export_attributes(self) -> dict[str, Any]:
        parent_attributes = super().get_export_attributes()
        attributes = [
            ("filter_type", "lbp"),
            ("filter_direction", "2d" if self.separate_slices else "3d"),
            ("distance", self.distance),
            ("lbp_method", self.lbp_method)
        ]
        parent_attributes.update(dict(attributes))

        return parent_attributes

    def parse_feature_names(self, x: None | pd.DataFrame) -> pd.DataFrame:
        x = super().parse_feature_names(x=x)
        feature_name_prefix = [
            "lbp",
            "2d" if self.separate_slices else "3d"
        ]

        # Don't add anything if the method is "default".
        if self.lbp_method == "variance":
            feature_name_prefix += ["var"]
        elif self.lbp_method == "rotation_invariant":
            feature_name_prefix += ["rot_invar"]

        feature_name_prefix += ["d" + str(self.distance)]

        if len(feature_name_prefix) > 0:
            feature_name_prefix = "_".join(feature_name_prefix)
            feature_name_prefix += "_"
            x.columns = feature_name_prefix + x.columns

        return x



import os
import pandas as pd
from pathlib import Path

from mirp._data_import.generic_file import ImageFile


def extract_image_parameters(
        image,
        sample_name: None | str | list[str] = None,
        image_name: None | str | list[str] = None,
        image_file_type: None | str = None,
        image_modality: None | str | list[str] = None,
        image_sub_folder: None | str = None,
        stack_images: str = "auto",
        write_dir: None | str | Path = None
) -> pd.DataFrame | None:
    """
    Extract parameters related to image acquisition and reconstruction from images. Not all metadata may
    be available.

    Parameters
    ----------
    image: Any
        A path to an image file, a path to a directory containing image files, a path to a config_data.xml
        file, a path to a csv file containing references to image files, a pandas.DataFrame containing references to
        image files, or a numpy.ndarray.

    sample_name: str or list of str, default: None
        Name of expected sample names. This is used to select specific image files. If None, no image files are
        filtered based on the corresponding sample name (if known).

    image_name: str, optional, default: None
        Pattern to match image files against. The matches are exact. Use wildcard symbols ("*") to
        match varying structures. The sample name (if part of the file name) can also be specified using "#". For
        example, image_name = '#_*_image' would find John_Doe in John_Doe_CT_image.nii or John_Doe_001_image.nii.
        File extensions do not need to be specified. If None, file names are not used for filtering files and
        setting sample names.

    image_file_type: {"dicom", "nifti", "nrrd", "numpy", "itk"}, optional, default: None
        The type of file that is expected. If None, the file type is not used for filtering files.
        "itk" comprises "nifti" and "nrrd" file types.

    image_modality: {"ct", "pet", "pt", "mri", "mr", "rtdose", "generic"}, optional, default: None
        The type of modality that is expected. If None, modality is not used for filtering files. Note that only
        DICOM files contain metadata concerning modality.

    image_sub_folder: str, optional, default: None
        Fixed directory substructure where image files are located. If None, the directory substructure is not used
        for filtering files.

    stack_images: {"auto", "yes", "no"}, optional, default: "str"
        If image files in the same directory cannot be assigned to different samples, and are 2D (slices) of the same
        size, they might belong to the same 3D image stack. "auto" will stack 2D numpy arrays, but not other file types.
        "yes" will stack all files that contain 2D images, that have the same dimensions, orientation and spacing,
        except for DICOM files. "no" will not stack any files. DICOM files ignore this argument, because their stacking
        can be determined from metadata.

    write_dir: str or Path, optional, default: None
        Directory to which the table with image acquisition and reconstruction metadata is written. Image metadata are
        written to ``image_metadata.csv`` in this directory.

    Returns
    -------
    pd.DataFrame | None
        The functions returns a table with metadata (``write_dir = None``) or nothing.
    """

    from mirp.data_import.import_image import import_image

    image_list = import_image(
        image=image,
        sample_name=sample_name,
        image_name=image_name,
        image_file_type=image_file_type,
        image_modality=image_modality,
        image_sub_folder=image_sub_folder,
        stack_images=stack_images
    )

    metadata = [_extract_image_parameters(ii, image) for ii, image in enumerate(image_list)]
    metadata = pd.DataFrame(metadata)

    if write_dir is not None:
        write_dir = os.path.normpath(write_dir)
        if not os.path.exists(write_dir):
            os.makedirs(write_dir)

        metadata.to_csv(
            path_or_buf=os.path.join(write_dir, "image_metadata.csv")
        )
    else:
        return metadata


def _extract_image_parameters(index: int, image: ImageFile) -> dict[str, str]:

    metadata = image.export_metadata()
    metadata.update({"image_index": index})

    return metadata



from mirp._images.mask_image import MaskImage
from mirp._data_import.utilities import flatten_list
from mirp._masks.base_mask import BaseMask


def alter_mask(
        masks: BaseMask | MaskImage | list[BaseMask],
        alteration_size: None | list[float] = None,
        alteration_method: None | str = None,
        max_erosion: None | float = 0.8
):
    """ Adapt roi size by growing or shrinking the roi """

    if alteration_size is None or alteration_method is None:
        return masks

    # Determine the return format.
    return_list = False
    if isinstance(masks, list):
        return_list = True
    else:
        masks = [masks]

    new_masks = []

    for mask in masks:
        for current_adapt_size in alteration_size:
            new_mask = mask.copy()
            if alteration_method == "distance" and current_adapt_size < 0.0:
                if isinstance(new_mask, BaseMask):
                    new_mask.roi.erode(
                        max_eroded_volume_fraction=max_erosion,
                        distance=current_adapt_size
                    )
                elif isinstance(new_mask, MaskImage):
                    new_mask.erode(
                        max_eroded_volume_fraction=max_erosion,
                        distance=current_adapt_size
                    )

            elif alteration_method == "distance" and current_adapt_size > 0.0:
                if isinstance(new_mask, BaseMask):
                    new_mask.roi.dilate(
                        distance=current_adapt_size
                    )
                elif isinstance(new_mask, MaskImage):
                    new_mask.dilate(
                        distance=current_adapt_size
                    )

            elif alteration_method == "fraction":
                if isinstance(new_mask, BaseMask):
                    new_mask.roi.fractional_volume_change(
                        fractional_change=current_adapt_size
                    )
                elif isinstance(new_mask, MaskImage):
                    new_mask.fractional_volume_change(
                        fractional_change=current_adapt_size
                    )

            if new_mask is not None:
                new_masks += [new_mask]

    new_masks = flatten_list(new_masks)
    if len(new_masks) == 0:
        return None

    if not return_list and len(new_masks) == 1:
        return new_masks[0]
    else:
        return new_masks



import warnings
import os

import numpy as np

from typing import Any
from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomFileRTDose(ImageDicomFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        return False

    def create(self):
        return self

    def _complete_image_orientation(self, force=False, frame_id=None):
        if self.image_orientation is None:
            # Load relevant metadata.
            self.load_metadata(limited=True)

            # This is orientation for x and y directions.
            orientation: list[float] = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x0037),
                tag_type="mult_float"
            )

            # First compute z-orientation.
            orientation += list(np.cross(orientation[0:3], orientation[3:6]))
            self.image_orientation = np.reshape(orientation[::-1], [3, 3], order="F")

    def _complete_image_spacing(self, force=False, frame_id=None):
        if self.image_spacing is None:

            # Load relevant metadata.
            self.load_metadata(limited=True)

            # Get pixel-spacing.
            spacing = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0030), tag_type="mult_float")

            # Get Grid Frame Offset Vector
            z_spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x3004, 0x000C),
                tag_type="mult_float"
            )

            if len(z_spacing) > 1:
                # DICOM file contains multiple frames.
                z_spacing = np.unique(np.diff(z_spacing))
                if len(z_spacing) > 1:
                    raise ValueError(
                        f"Spacing of radiation dose grid is inconsistent: {', '.join(z_spacing)}. "
                        f"[{self.describe_self()}]"
                    )
                z_spacing = z_spacing[0]
            else:
                # DICOM file contains only a single frame. Use a default 1.0 mm value.
                z_spacing = 1.0
                warnings.warn(
                    f"Radiation dose grid only contains a single frame (slice). A default frame spacing of 1.0 mm is "
                    f"assumed. Within-plane spacing is not affected. [{self.describe_self()}]"
                )

            if z_spacing < 0.0:
                self.image_orientation = np.matmul(
                    self.image_orientation, [[-1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
                )
                z_spacing = np.abs(z_spacing)

            spacing += [z_spacing]

            self.image_spacing = tuple(spacing[::-1])

    def _complete_image_dimensions(self, force=False):
        if self.image_dimension is None:

            # Load relevant metadata.
            self.load_metadata(limited=True)

            dimensions = tuple([
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0008), tag_type="int"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0010), tag_type="int"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0011), tag_type="int")
            ])

            self.image_dimension = dimensions

    def load_data(self, **kwargs):
        if self.image_data is not None:
            return self.image_data

        if self.file_path is not None and not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path} [{self.describe_self()}]"
            )

        if self.file_path is None:
            raise ValueError(f"A path to a file was expected, but not present. [{self.describe_self()}]")

        # Load metadata.
        self.load_metadata(include_image=True)
        image_data = self.image_metadata.pixel_array.astype(np.float32)

        # Update data with dose grid scaling
        dose_grid_scaling = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x3004, 0x000E),
            tag_type="float",
            default=1.0
        )
        image_data = image_data * dose_grid_scaling

        self.image_data = image_data

    def export_metadata(self, self_only: bool = False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Manufacturer
        manufacturer = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x0070),
            tag_type="str"
        )
        if manufacturer is not None:
            dcm_meta_data += [("manufacturer", manufacturer)]

        # Dose Units
        dose_units = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x3004, 0x0002),
            tag_type="str"
        )
        if dose_units is not None:
            dcm_meta_data += [("dose_units", dose_units)]

        # Dose Type
        dose_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x3004, 0x0004),
            tag_type="str"
        )
        if dose_type is not None:
            dcm_meta_data += [("dose_type", dose_type)]

        # Dose comment
        dose_comment = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x3004, 0x0006),
            tag_type="str"
        )
        if dose_comment is not None:
            dcm_meta_data += [("dose_comment", dose_comment)]

        # Dose summation type
        dose_summation_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x3004, 0x000a),
            tag_type="str"
        )
        if dose_summation_type is not None:
            dcm_meta_data += [("dose_summation_type", dose_summation_type)]

        metadata.update(dict(dcm_meta_data))
        return metadata



from typing import Any

import numpy as np

from mirp._images.generic_image import GenericImage
from mirp._images.mask_image import MaskImage
from mirp._masks.base_mask import BaseMask


def standard_image_process_checks(
        image: GenericImage,
        masks: None | BaseMask | MaskImage | list[BaseMask]
) -> tuple[GenericImage, None | list[BaseMask] | list[MaskImage], None | bool]:
    if masks is None:
        return image, None, None
    if isinstance(masks, list) and len(masks) == 0:
        return image, None, None

    # Determine the return format.
    return_list = False
    if isinstance(masks, list):
        return_list = True
    else:
        masks = [masks]

    if not isinstance(image, GenericImage):
        raise TypeError(
            f"The image argument is expected to be a GenericImage object, or inherit from it. Found: {type(image)}")

    if not all(isinstance(mask, BaseMask) or isinstance(mask, MaskImage) for mask in masks):
        raise TypeError(
            f"The masks argument is expected to be a BaseMask or MaskImage object, or a list thereof.")

    return image, masks, return_list


def set_intensity_range(
        image: GenericImage,
        mask: None | MaskImage = None,
        intensity_range: tuple[Any, Any] | None = None
) -> tuple[float, float]:
    if intensity_range is not None and not np.any(np.isnan(intensity_range)):
        return intensity_range

    if mask is None or mask.is_empty() or mask.is_empty_mask():
        mask_data = np.ones(image.image_dimension, dtype=bool)
    else:
        mask_data = mask.get_voxel_grid()

    # Make intensity range mutable.
    if intensity_range is None:
        intensity_range = [np.nan, np.nan]
    else:
        intensity_range = list(intensity_range)

    if np.isnan(intensity_range[0]):
        intensity_range[0] = np.min(image.get_voxel_grid()[mask_data])
    if np.isnan(intensity_range[1]):
        intensity_range[1] = np.max(image.get_voxel_grid()[mask_data])

    return tuple(intensity_range)


def extend_intensity_range(
        intensity_range: tuple[Any, Any],
        extend_fraction=0.1
) -> None | tuple[Any, Any]:
    if intensity_range is None or np.any(np.isnan(intensity_range)):
        return intensity_range

    if extend_fraction <= 0.0:
        return intensity_range

    # Add 10% range outside the grey level range
    extension = 0.1 * (intensity_range[1] - intensity_range[0])
    intensity_range = list(intensity_range)
    intensity_range[0] -= extension
    intensity_range[1] += extension

    return tuple(intensity_range)



import numpy as np
from typing import Any

from mirp._images.generic_image import GenericImage


def create_tissue_mask(
        image: GenericImage,
        mask_type: None | str = None,
        mask_intensity_range: None | tuple[Any, Any] = None
) -> np.ndarray:

    if mask_type is None or mask_type == "none":
        # The entire image is the tissue mask.
        mask = np.ones(image.image_dimension, dtype=bool)

    elif mask_type == "range":
        if mask_intensity_range is None:
            mask_intensity_range = [np.nan, np.nan]
        else:
            mask_intensity_range = list(mask_intensity_range)

        # The intensity range provided forms the mask range.
        if np.isnan(mask_intensity_range[0]):
            mask_intensity_range[0] = np.min(image.get_voxel_grid())
        if np.isnan(mask_intensity_range[1]):
            mask_intensity_range[1] = np.max(image.get_voxel_grid())

        voxel_grid = image.get_voxel_grid()
        mask = np.logical_and(voxel_grid >= mask_intensity_range[0], voxel_grid <= mask_intensity_range[1])

    elif mask_type == "relative_range":
        from skimage.morphology import binary_opening

        # The relative intensity range provided forms the mask range. This means that we need to convert the relative
        # range to the range present in the image.
        if mask_intensity_range is None:
            mask_intensity_range = [np.nan, np.nan]
        else:
            mask_intensity_range = list(mask_intensity_range)

        if np.isnan(mask_intensity_range[0]):
            mask_intensity_range[0] = 0.0
        if np.isnan(mask_intensity_range[1]):
            mask_intensity_range[1] = 1.0

        voxel_grid = image.get_voxel_grid()
        intensity_range = [np.min(voxel_grid), np.max(voxel_grid)]

        # Convert relative range to the image intensities
        tissue_range = [
            intensity_range[0] + mask_intensity_range[0] * (intensity_range[1] - intensity_range[0]),
            intensity_range[0] + mask_intensity_range[1] * (intensity_range[1] - intensity_range[0])
        ]

        mask = np.logical_and(voxel_grid >= tissue_range[0], voxel_grid <= tissue_range[1])

        # Perform binary closing to smooth the mask.
        mask = binary_opening(mask)

    else:
        raise ValueError(f"The tissue_mask_type configuration parameter is expected to be one of none, range, "
                         f"or relative_range. Encountered: {mask_type}")

    # Check that masks are not completely empty.
    if np.all(np.logical_not(mask)):
        mask = np.ones(image.image_dimension, dtype=bool)

    return mask



import numpy as np

from mirp._masks.base_mask import BaseMask


def split_masks(
        masks: None | BaseMask | list[BaseMask],
        boundary_sizes: None | list[float] = None,
        max_erosion: None | float = 0.8
):
    if boundary_sizes is None or len(boundary_sizes) == 0 or \
            all(boundary_size == 0.0 for boundary_size in boundary_sizes):
        return masks

    if masks is None:
        return None

    # Determine the return format.
    if not isinstance(masks, list):
        masks = [masks]

    masks: list[BaseMask] = masks
    new_masks = []

    # Iterate over masks.
    for mask in masks:
        # Store original.
        new_masks += [mask]

        for boundary_size in boundary_sizes:
            if boundary_size == 0.0:
                continue

            bulk_mask = mask.copy()
            bulk_mask.roi_name += "_bulk_" + str(boundary_size)

            boundary_mask = mask.copy()
            boundary_mask.roi_name += "_boundary_" + str(boundary_size)

            bulk_mask.erode(
                distance=boundary_size,
                max_eroded_volume_fraction=max_erosion
            )

            boundary_mask.roi.set_voxel_grid(voxel_grid=np.logical_xor(
                mask.roi.get_voxel_grid(),
                bulk_mask.roi.get_voxel_grid()
            ))

            if bulk_mask.roi.is_empty_mask() or boundary_mask.roi.is_empty_mask():
                continue

            new_masks += [bulk_mask, boundary_mask]

    return new_masks



from mirp._image_processing.utilities import standard_image_process_checks
from mirp._images.generic_image import GenericImage
from mirp._images.mask_image import MaskImage
from mirp._data_import.utilities import flatten_list
from mirp._masks.base_mask import BaseMask


def randomise_mask(
        image: GenericImage,
        masks: BaseMask | MaskImage | list[BaseMask],
        boundary: float = 25.0,
        repetitions: int = 1
):
    image, masks, return_list = standard_image_process_checks(image=image, masks=masks)
    if return_list is None:
        return None

    new_masks = []
    for mask in masks:
        if isinstance(mask, MaskImage):
            randomised_masks = mask.randomise_mask(
                image=image,
                boundary=boundary,
                repetitions=repetitions
            )

            for randomised_mask in randomised_masks:
                if randomised_mask is None:
                    continue
                new_masks += [randomised_mask]

        elif isinstance(mask, BaseMask):
            randomised_masks = mask.roi.randomise_mask(
                image=image,
                boundary=boundary,
                repetitions=repetitions,
                intensity_range=mask.intensity_range
            )

            for randomised_mask in randomised_masks:
                if randomised_mask is None:
                    continue
                new_mask = mask.copy(drop_image=True)
                new_mask.roi = randomised_mask
                new_masks += [new_mask]
        else:
            raise TypeError("The masks attribute is expected to be MaskImage and BaseMask")

    new_masks = flatten_list(new_masks)
    if len(new_masks) == 0:
        return None

    if not return_list and repetitions == 1:
        return new_masks[0]
    else:
        return new_masks



from mirp.data_import.import_image import import_image
from mirp.data_import.import_mask import import_mask
from mirp._data_import.generic_file import ImageFile, MaskFile, MaskFullImage
from mirp._data_import.dicom_file import ImageDicomFile, MaskDicomFile
from mirp._data_import.dicom_file_stack import ImageDicomFileStack
from mirp.utilities.utilities import random_string


def import_image_and_mask(
        image,
        mask=None,
        sample_name: None | str | list[str] = None,
        image_name: None | str | list[str] = None,
        image_file_type: None | str = None,
        image_modality: None | str | list[str] = None,
        image_sub_folder: None | str = None,
        mask_name: None | str | list[str] = None,
        mask_file_type: None | str = None,
        mask_modality: None | str | list[str] = None,
        mask_sub_folder: None | str = None,
        roi_name: None | str | list[str] | dict[str, str] = None,
        association_strategy: None | str | list[str] = None,
        stack_images: str = "auto",
        stack_masks: str = "auto"
) -> list[ImageFile]:
    """
    Creates and curates references to image and mask files. This function is usually called internally by other
    functions such as :func:`~mirp.extractFeaturesAndImages.extract_features`.

    Parameters
    ----------
    image: Any
        A path to an image file, a path to a directory containing image files, a path to a config_data.xml
        file, a path to a csv file containing references to image files, a pandas.DataFrame containing references to
        image files, or a numpy.ndarray.

    mask: Any
        A path to a mask file, a path to a directory containing mask files, a path to a config_data.xml
        file, a path to a csv file containing references to mask files, a pandas.DataFrame containing references to
        mask files, or a numpy.ndarray.

    sample_name: str or list of str, default: None
        Name of expected sample names. This is used to select specific image files. If None, no image files are
        filtered based on the corresponding sample name (if known).

    image_name: str, optional, default: None
        Pattern to match image files against. The matches are exact. Use wildcard symbols ("*") to
        match varying structures. The sample name (if part of the file name) can also be specified using "#". For
        example, image_name = '#_*_image' would find John_Doe in John_Doe_CT_image.nii or John_Doe_001_image.nii.
        File extensions do not need to be specified. If None, file names are not used for filtering files and
        setting sample names.

    image_file_type: {"dicom", "nifti", "nrrd", "numpy", "itk"}, optional, default: None
        The type of file that is expected. If None, the file type is not used for filtering files.
        "itk" comprises "nifti" and "nrrd" file types.

    image_modality: {"ct", "pet", "pt", "mri", "mr", "rtdose", "generic"}, optional, default: None
        The type of modality that is expected. If None, modality is not used for filtering files. Note that only
        DICOM files contain metadata concerning modality.

    image_sub_folder: str, optional, default: None
        Fixed directory substructure where image files are located. If None, the directory substructure is not used
        for filtering files.

    mask_name: str or list of str, optional, default: None
        Pattern to match mask files against. The matches are exact. Use wildcard symbols ("*") to match varying
        structures. The sample name (if part of the file name) can also be specified using "#". For example,
        mask_name = '#_*_mask' would find John_Doe in John_Doe_CT_mask.nii or John_Doe_001_mask.nii. File extensions
        do not need to be specified. If None, file names are not used for filtering files and setting sample names.

    mask_file_type: {"dicom", "nifti", "nrrd", "numpy", "itk"}, optional, default: None
        The type of file that is expected. If None, the file type is not used for filtering files.
        "itk" comprises "nifti" and "nrrd" file types.

    mask_modality: {"rtstruct", "seg", "generic_mask"}, optional, default: None
        The type of modality that is expected. If None, modality is not used for filtering files.
        Note that only DICOM files contain metadata concerning modality. Masks from non-DICOM files are considered to
        be "generic_mask".

    mask_sub_folder: str, optional, default: None
        Fixed directory substructure where mask files are located. If None, the directory substructure is not used for
        filtering files.

    roi_name: str or list of str or dict, optional, default: None
        Name of the regions of interest that should be assessed.

    association_strategy: {"frame_of_reference", "sample_name", "file_distance", "file_name_similarity",  "list_order", "position", "single_image"}
        The preferred strategy for associating images and masks. File association is preferably done using frame of
        reference UIDs (DICOM), or sample name (NIfTI, numpy). Other options are relatively frail, except for
        `list_order` which may be applicable when a list with images and a list with masks is provided and both lists
        are of equal length.

    stack_images: {"auto", "yes", "no"}, optional, default: "str"
        If image files in the same directory cannot be assigned to different samples, and are 2D (slices) of the same
        size, they might belong to the same 3D image stack. "auto" will stack 2D numpy arrays, but not other file types.
        "yes" will stack all files that contain 2D images, that have the same dimensions, orientation and spacing,
        except for DICOM files. "no" will not stack any files. DICOM files ignore this argument, because their stacking
        can be determined from metadata.

    stack_masks: {"auto", "yes", "no"}, optional, default: "str"
        If mask files in the same directory cannot be assigned to different samples, and are 2D (slices) of the same
        size, they might belong to the same 3D mask stack. "auto" will stack 2D numpy arrays, but not other file
        types. "yes" will stack all files that contain 2D images, that have the same dimensions, orientation and
        spacing, except for DICOM files. "no" will not stack any files. DICOM files ignore this argument,
        because their stacking can be determined from metadata.

    Returns
    -------
    list[ImageFile]
        The functions returns a list of ImageFile objects, if any were found with the specified filters.
    """

    # If mask = None, this can mean several things. Here we check that mask should not be interpreted as having the
    # same meaning as image.
    if mask is None and (
            mask_name is not None or mask_sub_folder is not None or mask_modality is not None or
            mask_file_type is not None or roi_name is not None
    ):
        mask = image

    elif mask is None and isinstance(image, str) and image.endswith(".xml"):
        mask = image

    # Generate list of images.
    image_list = import_image(
        image,
        sample_name=sample_name,
        image_name=image_name,
        image_file_type=image_file_type,
        image_modality=image_modality,
        image_sub_folder=image_sub_folder,
        stack_images=stack_images
    )

    if len(image_list) == 0:
        raise ValueError(f"No images were found. Possible reasons are lack of images with the preferred modality.")

    # Remove metadata (if any) - anything relevant can be loaded later.
    for image in image_list:
        image.remove_metadata()

    if mask is not None:
        # Generate list of masks from mask.
        mask_list = import_mask(
            mask,
            sample_name=sample_name,
            mask_name=mask_name,
            mask_file_type=mask_file_type,
            mask_modality=mask_modality,
            mask_sub_folder=mask_sub_folder,
            stack_masks=stack_masks,
            roi_name=roi_name
        )

        if len(mask_list) == 0:
            raise ValueError(f"No masks were found. Possible reasons are lack of masks with the preferred modality.")

        # Remove metadata (if any) - anything relevant can be loaded later.
        for mask in mask_list:
            mask.remove_metadata()

        # Determine association strategy, if this is unset.
        possible_association_strategy = set_association_strategy(
            image_list=image_list,
            mask_list=mask_list
        )

        if association_strategy is None:
            association_strategy = possible_association_strategy
        elif isinstance(association_strategy, str):
            association_strategy = [association_strategy]

        if not isinstance(association_strategy, set):
            association_strategy = set(association_strategy)

        # Test association strategy.
        unavailable_strategy = association_strategy - possible_association_strategy
        if len(unavailable_strategy) > 0:
            raise ValueError(
                f"One or more strategies for associating images and masks are not available for the provided image and "
                f"mask set: {', '.join(list(unavailable_strategy))}. Only the following strategies are available: "
                f"{'. '.join(list(possible_association_strategy))}"
            )

        if len(possible_association_strategy) == 0:
            raise ValueError(
                f"No strategies for associating images and masks are available, indicating that there is no clear way to "
                f"establish an association."
            )

        # Start association.
        if association_strategy == {"list_order"}:
            # If only the list_order strategy is available, use this.
            for ii, image in enumerate(image_list):
                image.associated_masks = [mask_list[ii]]

        elif association_strategy == {"single_image"}:
            # If single_image is the only strategy, use this.
            image_list[0].associated_masks = mask_list

        else:
            for ii, image in enumerate(image_list):
                image.associate_with_mask(
                    mask_list=mask_list,
                    association_strategy=association_strategy
                )

            if all(image.associated_masks is None for image in image_list):
                if "single_image" in association_strategy:
                    image_list[0].associated_masks = mask_list
                elif "list_order" in association_strategy:
                    for ii, image in enumerate(image_list):
                        image.associated_masks = [mask_list[ii]]

    else:
        # Generate full masks.
        for image in image_list:
            image.associated_masks = [MaskFullImage()]

    # In case we are not working with file-based images, ensure that we are working with deep copies from this point -
    # we don't want to propagate changes to masks and images by reference.
    image_list = [image if image.on_file_system() else image.copy() for image in image_list]

    # Set sample names. First we check if all sample names are missing.
    sample_name_set = all(image.sample_name is not None for image in image_list)

    # If a single string is present, convert to list of str.
    if not sample_name_set and isinstance(sample_name, str):
        sample_name = [sample_name]

    # Use provided sample names.
    if not sample_name_set and isinstance(sample_name, list) and len(sample_name) == len(image_list):
        for ii, image in enumerate(image_list):
            image.set_sample_name(sample_name=sample_name[ii])
            if image.associated_masks is not None:
                for mask in image.associated_masks:
                    mask.set_sample_name(sample_name=sample_name[ii])
        sample_name_set = True

    # Infer sample name from file name.
    if not sample_name_set and all(image.file_name is not None for image in image_list):
        # Check that this results in unique sample names.
        file_sample_names = [image.get_sample_name_from_file(must_succeed=True) for image in image_list]
        if len(set(file_sample_names)) == len(image_list):
            for ii, image in enumerate(image_list):
                image.set_sample_name(sample_name=file_sample_names[ii])

                if image.associated_masks is not None:
                    for mask in image.associated_masks:
                        mask.set_sample_name(sample_name=file_sample_names[ii])

            sample_name_set = True

    # Infer sample name from folder structure.
    if not sample_name_set and all(image.dir_path is not None for image in image_list):
        # Check that this results in unique sample names.
        folder_sample_names = [image.get_sample_name_from_folder(sub_folder=image_sub_folder) for image in image_list]
        if len(set(folder_sample_names)) == len(image_list):
            for ii, image in enumerate(image_list):
                image.set_sample_name(sample_name=folder_sample_names[ii])

                if image.associated_masks is not None:
                    for mask in image.associated_masks:
                        mask.set_sample_name(sample_name=folder_sample_names[ii])
            sample_name_set = True

    # Then set any sample names for images that still miss them.
    if any(image.sample_name is None for image in image_list):
        for ii, image in enumerate(image_list):
            if image.sample_name is None:
                generated_sample_name = str(ii + 1) + "_" + random_string(16)
                image.set_sample_name(sample_name=generated_sample_name)
                if image.associated_masks is not None:
                    for mask in image.associated_masks:
                        mask.set_sample_name(sample_name=generated_sample_name)

    return image_list


def set_association_strategy(
        image_list: list[ImageFile] | list[ImageDicomFile],
        mask_list: list[MaskFile] | list[MaskDicomFile]
) -> set[str]:
    # Association strategy is set by a process of elimination.
    possible_strategies = {
        "frame_of_reference", "sample_name", "file_distance", "file_name_similarity",  "list_order", "position",
        "single_image"
    }

    # Check that images and masks are available
    if len(mask_list) == 0 or len(image_list) == 0:
        return set([])

    # Check if association by list order is possible.
    if len(image_list) != len(mask_list):
        possible_strategies.remove("list_order")

    # Check that association with a single image is possible.
    if len(image_list) > 1:
        possible_strategies.remove("single_image")

    # Check if association by frame of reference UID is possible.
    if (any(isinstance(image, ImageDicomFile) or isinstance(image, ImageDicomFileStack) for image in image_list) and
            any(isinstance(mask, MaskDicomFile) for mask in mask_list)):
        dcm_image_list: list[ImageDicomFile | ImageDicomFileStack] = [
            image for image in image_list
            if isinstance(image, ImageDicomFile) or isinstance(image, ImageDicomFileStack)]
        dcm_mask_list: list[MaskDicomFile] = [mask for mask in mask_list if isinstance(mask, MaskDicomFile)]

        # If frame of reference UIDs are completely absent.
        if all(image.frame_of_reference_uid is None for image in dcm_image_list) or \
                all(mask.frame_of_reference_uid is None for mask in dcm_mask_list):
            possible_strategies.remove("frame_of_reference")

    else:
        possible_strategies.remove("frame_of_reference")

    # Check if association by sample name is possible.
    if all(image.sample_name is None for image in image_list) or all(mask.sample_name is None for mask in mask_list):
        possible_strategies.remove("sample_name")

    # Check if file_distance is possible. If directory are absent or singular, file distance cannot be used for
    # association.
    image_dir_path = set(image.get_dir_path() for image in image_list) - {None}
    mask_dir_path = set(mask.get_dir_path() for mask in mask_list) - {None}
    if len(image_dir_path) == 0 or len(mask_dir_path) <= 1:
        possible_strategies.remove("file_distance")

    # Check if file_name_similarity is possible. If file names are absent, this is not possible.
    if all(image.get_file_name() is None for image in image_list) or all(mask.get_file_name() is None for mask in mask_list):
        possible_strategies.remove("file_name_similarity")

    # Check if position can be used.
    if all(image.get_image_origin() is None for image in image_list) or all(mask.get_image_origin() is None for mask in mask_list):
        possible_strategies.remove("position")
    else:
        image_position_data = set([
            image.get_image_origin(as_str=True) + image.get_image_spacing(as_str=True) +
            image.get_image_dimension(as_str=True) + image.get_image_orientation(as_str=True)
            for image in image_list if image.get_image_origin() is not None
        ])
        mask_position_data = set([
            mask.get_image_origin(as_str=True) + mask.get_image_spacing(as_str=True) +
            mask.get_image_dimension(as_str=True) + mask.get_image_orientation(as_str=True)
            for mask in mask_list if mask.get_image_origin() is not None
        ])

        # Check that there are more
        if len(image_position_data) <= 1 or len(mask_position_data) <= 1:
            possible_strategies.remove("position")

    return possible_strategies



from typing import Any
from dataclasses import dataclass
from mirp.settings.utilities import setting_def


@dataclass
class ImagePerturbationSettingsClass:
    """
    Parameters related to image and mask perturbation / augmentation. By default images and masks are not perturbed or
    augmented.

    Parameters
    ----------

    crop_around_roi: bool, optional, default: False
        Determines whether the image may be cropped around the regions of interest. Setting
        this to True may speed up computation and save memory.

    crop_distance: float, optional, default: 150.0
        Physical distance around the mask that should be maintained when cropping the image. When using convolutional
        kernels for filtering an image, we recommend to leave some distance to prevent boundary effects. A crop
        distance of 0.0 crops the image tightly around the mask.

    perturbation_noise_repetitions: int, optional, default: 0
        Number of repetitions where noise is randomly added to the image. A value of 0 means that no noise will be
        added.

    perturbation_noise_level: float, optional, default: None
        Set the noise level in intensity units. This determines the width of the normal distribution used to generate
        random noise. If None (default), noise is determined from the image itself.

    perturbation_rotation_angles: float or list of float, optional, default: 0.0
        Angles (in degrees) over which the image and mask are rotated. This rotation is only in the x-y (axial)
        plane. Multiple angles can be provided to create images with different rotations.

    perturbation_translation_fraction: float or list of float, optional, default: 0.0
        Sub-voxel translation distance fractions of the interpolation grid. This forces the interpolation grid to
        shift slightly and interpolate at different points. Multiple values can be provided. All values should be
        between 0.0 and 1.0.

    perturbation_roi_adapt_type: {"fraction", "distance"}, optional, default: "distance"
        Determines how the mask is grown or shrunk. Can be either "fraction" or "distance". "fraction" is used to
        grow or shrink the mask by a certain fraction (see the ``perturbation_roi_adapt_size`` parameter).
        "distance" is used to grow or shrink the mask by a certain physical distance, defined using the
        ``perturbation_roi_adapt_size`` parameter.

    perturbation_roi_adapt_size: float or list of float, optional, default: 0.0
        Determines the extent of growth/shrinkage of the ROI mask. The use of this parameter depends on the
        growth/shrinkage type (``perturbation_roi_adapt_type``), For "distance", this parameter defines
        growth/shrinkage in physical units, typically mm. For "fraction", this parameter defines growth/shrinkage in
        volume fraction (e.g. a value of 0.2 grows the mask by 20%). For either type, positive values indicate growing
        the mask, whereas negative values indicate its shrinkage. Multiple values can be provided to perturb the
        volume of the mask.

    perturbation_roi_adapt_max_erosion: float, optional, default: 0.8
        Limits shrinkage of the mask by distance-based adaptations to avoid forming empty masks. Defined as fraction of
        the original volume, e.g. a value of 0.8 prevents shrinking the mask below 80% of its original volume. Only
        used when ``perturbation_roi_adapt_type=="distance"``.

    perturbation_randomise_roi_repetitions: int, optional, default: 0.0
        Number of repetitions where the mask is randomised using supervoxel-based randomisation.

    roi_split_boundary_size: float or list of float, optional, default: 0.0
        Width of the rim used for splitting the mask into bulk and rim masks, in physical dimensions. Multiple values
        can be provided to generate rims of different widths.

    roi_split_max_erosion: float, optional, default: 0.6
        Determines the minimum volume of the bulk mask when splitting the original mask into bulk and rim sections.
        Fraction of the original volume, e.g. 0.6 means that the bulk contains at least 60% of the original mask.

    **kwargs: dict, optional
        Unused keyword arguments.
    """

    def __init__(
            self,
            crop_around_roi: bool = False,
            crop_distance: float = 150.0,
            perturbation_noise_repetitions: int = 0,
            perturbation_noise_level: None | float = None,
            perturbation_rotation_angles: None | float | list[float] = 0.0,
            perturbation_translation_fraction: None | float | list[float] = 0.0,
            perturbation_roi_adapt_type: str = "distance",
            perturbation_roi_adapt_size: None | float | list[float] = 0.0,
            perturbation_roi_adapt_max_erosion: float = 0.8,
            perturbation_randomise_roi_repetitions: int = 0,
            roi_split_boundary_size: None | float | list[float] = 0.0,
            roi_split_max_erosion: float = 0.6,
            **kwargs
    ):

        # Set crop_around_roi
        self.crop_around_roi: bool = crop_around_roi

        # Check that crop distance is not negative.
        if crop_distance < 0.0 and crop_around_roi:
            raise ValueError(f"The cropping distance cannot be negative. Found: {crop_distance}")

        # Set crop_distance.
        self.crop_distance: float = crop_distance

        # Check that noise repetitions is not negative.
        perturbation_noise_repetitions = int(perturbation_noise_repetitions)
        if perturbation_noise_repetitions < 0:
            raise ValueError(f"The number of repetitions where noise is added to the image cannot be negative. Found: {perturbation_noise_repetitions}")

        # Set noise repetitions.
        self.add_noise: bool = perturbation_noise_repetitions > 0
        self.noise_repetitions: int = perturbation_noise_repetitions

        # Check noise level.
        if perturbation_noise_level is not None:
            if perturbation_noise_level < 0.0:
                raise ValueError(f"The noise level cannot be negative. Found: {perturbation_noise_level}")

        # Set noise level.
        self.noise_level: None | float = perturbation_noise_level

        # Convert perturbation_rotation_angles to list, if necessary.
        if not isinstance(perturbation_rotation_angles, list):
            perturbation_rotation_angles = [perturbation_rotation_angles]

        # Check that the rotation angles are floating points.
        if not all(isinstance(ii, float) for ii in perturbation_rotation_angles):
            raise TypeError(f"Not all values for perturbation_rotation_angles are floating point values.")

        # Set rotation_angles.
        self.rotation_angles: list[float] = perturbation_rotation_angles

        # Convert perturbation_translation_fraction to list, if necessary.
        if not isinstance(perturbation_translation_fraction, list):
            perturbation_translation_fraction = [perturbation_translation_fraction]

        # Check that the translation fractions are floating points.
        if not all(isinstance(ii, float) for ii in perturbation_translation_fraction):
            raise TypeError(f"Not all values for perturbation_translation_fraction are floating point values.")

        # Check that the translation fractions lie between 0.0 and 1.0.
        if not all(0.0 <= ii < 1.0 for ii in perturbation_translation_fraction):
            raise ValueError(
                "Not all values for perturbation_translation_fraction lie between 0.0 and 1.0, not including 1.0."
            )

        # Set translation_fraction.
        self.translation_fraction: list[float] = perturbation_translation_fraction

        # Check roi adaptation type.
        if perturbation_roi_adapt_type not in ["distance", "fraction"]:
            raise ValueError(f"The perturbation ROI adaptation type should be one of 'distance' or 'fraction'. Found: {perturbation_roi_adapt_type}")

        # Set roi_adapt_type
        self.roi_adapt_type: str = perturbation_roi_adapt_type

        # Convert to perturbation_roi_adapt_size to list.
        if not isinstance(perturbation_roi_adapt_size, list):
            perturbation_roi_adapt_size = [perturbation_roi_adapt_size]

        # Check that the adapt sizes are floating points.
        if not all(isinstance(ii, float) for ii in perturbation_roi_adapt_size):
            raise TypeError(f"Not all values for perturbation_roi_adapt_size are floating point values.")

        # Check that values do not go below 0.
        if perturbation_roi_adapt_type == "fraction" and any([ii <= -1.0 for ii in perturbation_roi_adapt_size]):
            raise ValueError("All values for perturbation_roi_adapt_size should be greater than -1.0. However, "
                             "one or more values were less.")

        # Set roi_adapt_size
        self.roi_adapt_size: list[float] = perturbation_roi_adapt_size

        # Check that perturbation_roi_adapt_max_erosion is between 0.0 and 1.0.
        if not 0.0 <= perturbation_roi_adapt_max_erosion <= 1.0:
            raise ValueError(f"The perturbation_roi_adapt_max_erosion parameter must have a value between 0.0 and "
                             f"1.0. Found: {perturbation_roi_adapt_max_erosion}")

        # Set max volume erosion.
        self.max_volume_erosion: float = perturbation_roi_adapt_max_erosion

        # Check that ROI randomisation representation is not negative.
        perturbation_randomise_roi_repetitions = int(perturbation_randomise_roi_repetitions)
        if perturbation_randomise_roi_repetitions < 0:
            raise ValueError(
                f"The number of repetitions where the ROI mask is randomised cannot be negative. Found: "
                f"{perturbation_randomise_roi_repetitions}")

        # Set ROI mask randomisation repetitions.
        self.randomise_roi: bool = perturbation_randomise_roi_repetitions > 0
        self.roi_random_rep: int = perturbation_randomise_roi_repetitions

        # Check that roi_split_max_erosion is between 0.0 and 1.0.
        if not 0.0 <= roi_split_max_erosion <= 1.0:
            raise ValueError(f"The roi_split_max_erosion parameter must have a value between 0.0 and "
                             f"1.0. Found: {roi_split_max_erosion}")

        # Division of roi into bulk and boundary
        self.max_bulk_volume_erosion: float = roi_split_max_erosion

        # Convert roi_split_boundary_size to list, if necessary.
        if not isinstance(roi_split_boundary_size, list):
            roi_split_boundary_size = [roi_split_boundary_size]

        # Check that the translation fractions are floating points.
        if not all(isinstance(ii, float) for ii in roi_split_boundary_size):
            raise TypeError(f"Not all values for roi_split_boundary_size are floating point values.")

        # Check that the translation fractions lie between 0.0 and 1.0.
        if not all(ii >= 0.0 for ii in roi_split_boundary_size):
            raise ValueError("Not all values for roi_split_boundary_size are positive.")

        # Set roi_boundary_size.
        self.roi_boundary_size: list[float] = roi_split_boundary_size

        # Initially local variables
        self.translate_x: None | float = None
        self.translate_y: None | float = None
        self.translate_z: None | float = None


def get_perturbation_settings() -> list[dict[str, Any]]:
    return [
        setting_def("crop_around_roi", "bool", xml_key=["crop_around_roi", "resect"], test=True),
        setting_def("crop_distance", "float", test=10.0),
        setting_def(
            "perturbation_noise_repetitions", "int", xml_key="noise_repetitions",
            class_key="noise_repetitions", test=10
        ),
        setting_def(
            "perturbation_noise_level", "float", xml_key="noise_level", class_key="noise_level", test=0.75
        ),
        setting_def(
            "perturbation_rotation_angles", "float", to_list=True, xml_key=["rotation_angles", "rot_angles"],
            class_key="rotation_angles", test=[-33.0, 33.0]
        ),
        setting_def(
            "perturbation_translation_fraction", "float", to_list=True,
            xml_key=["translation_fraction", "translate_frac"], class_key="translation_fraction", test=[0.25, 0.75]
        ),
        setting_def(
            "perturbation_roi_adapt_type", "str", xml_key="roi_adapt_type", class_key="roi_adapt_type",
            test="fraction"
        ),
        setting_def(
            "perturbation_roi_adapt_size", "float", to_list=True, xml_key="roi_adapt_size",
            class_key="roi_adapt_size", test=[0.8, 1.0, 1.2]
        ),
        setting_def(
            "perturbation_roi_adapt_max_erosion", "float", xml_key=["roi_adapt_max_erosion", "eroded_vol_fract"],
            class_key="max_volume_erosion", test=0.2
        ),
        setting_def(
            "perturbation_randomise_roi_repetitions", "int", xml_key="roi_random_rep",
            class_key="roi_random_rep", test=100
        ),
        setting_def(
            "roi_split_boundary_size", "float", to_list=True, xml_key="roi_boundary_size",
            class_key="roi_boundary_size", test=[2.0, 5.0]
        ),
        setting_def(
            "roi_split_max_erosion", "float", xml_key=["roi_split_max_erosion", "bulk_min_vol_fract"],
            class_key="max_bulk_volume_erosion", test=0.2
        )
    ]




import numpy as np


def real_ndim(x: np.ndarray):
    """
    Determine dimensionality of an array based on its shape. This is unlike the ndim method, which shows the number
    of dimensions based on the length of shape.

    :param x: A numpy array.
    :return: The number of dimensions.
    """
    # Compute the number of dimensions by getting the number of dimensions that have length 0 or 1.
    number_dims = x.ndim - sum([current_dim <= 1 for current_dim in x.shape])

    # If all dimensions are 1, dimensionality is technically 0, but 1 in practice.
    if number_dims < 1:
        number_dims = 1

    return number_dims


def random_string(k: int) -> str:
    import string
    from random import choices

    character_list = list(string.ascii_uppercase + string.digits)
    return "".join(choices(character_list, k=k))


def zero_if_close(x: np.ndarray, tol: float = 100.0):
    f =  np.finfo(x.dtype)
    x[np.abs(x) < f.eps * tol] = 0.0
    return x
